from http import HTTPStatus
from typing import Any, Union

import httpx

from ... import errors
from ...client import Client
from ...models.agent import Agent
from ...models.error import Error
from ...types import Response


def _get_kwargs() -> dict[str, Any]:
    _kwargs: dict[str, Any] = {
        "method": "get",
        "url": "/agents",
    }

    return _kwargs


def _parse_response(
    *, client: Client, response: httpx.Response
) -> Union[Error, list["Agent"]] | None:
    if response.status_code == 200:
        response_200 = []
        _response_200 = response.json()
        for response_200_item_data in _response_200:
            response_200_item = Agent.from_dict(response_200_item_data)

            response_200.append(response_200_item)

        return response_200
    if response.status_code == 401:
        response_401 = Error.from_dict(response.json())

        return response_401
    if response.status_code == 403:
        response_403 = Error.from_dict(response.json())

        return response_403
    if response.status_code == 500:
        response_500 = Error.from_dict(response.json())

        return response_500
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Client, response: httpx.Response
) -> Response[Union[Error, list["Agent"]]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: Client,
) -> Response[Union[Error, list["Agent"]]]:
    """List all agents

     Returns all AI agents deployed in the workspace. Each agent includes its deployment status, runtime
    configuration, and global inference endpoint URL.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Error, list['Agent']]]
    """

    kwargs = _get_kwargs()

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: Client,
) -> Union[Error, list["Agent"]] | None:
    """List all agents

     Returns all AI agents deployed in the workspace. Each agent includes its deployment status, runtime
    configuration, and global inference endpoint URL.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Error, list['Agent']]
    """

    return sync_detailed(
        client=client,
    ).parsed


async def asyncio_detailed(
    *,
    client: Client,
) -> Response[Union[Error, list["Agent"]]]:
    """List all agents

     Returns all AI agents deployed in the workspace. Each agent includes its deployment status, runtime
    configuration, and global inference endpoint URL.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Error, list['Agent']]]
    """

    kwargs = _get_kwargs()

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: Client,
) -> Union[Error, list["Agent"]] | None:
    """List all agents

     Returns all AI agents deployed in the workspace. Each agent includes its deployment status, runtime
    configuration, and global inference endpoint URL.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Error, list['Agent']]
    """

    return (
        await asyncio_detailed(
            client=client,
        )
    ).parsed



from http import HTTPStatus
from typing import Any

import httpx

from ... import errors
from ...client import Client
from ...models.cleanup_images_response_200 import CleanupImagesResponse200
from ...types import Response


def _get_kwargs() -> dict[str, Any]:
    _kwargs: dict[str, Any] = {
        "method": "delete",
        "url": "/images",
    }

    return _kwargs


def _parse_response(*, client: Client, response: httpx.Response) -> CleanupImagesResponse200 | None:
    if response.status_code == 200:
        response_200 = CleanupImagesResponse200.from_dict(response.json())

        return response_200
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Client, response: httpx.Response
) -> Response[CleanupImagesResponse200]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: Client,
) -> Response[CleanupImagesResponse200]:
    """Cleanup unused container images

     Cleans up unused container images in the workspace registry. Only removes images that are not
    currently referenced by any active agent, function, sandbox, or job deployment.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[CleanupImagesResponse200]
    """

    kwargs = _get_kwargs()

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: Client,
) -> CleanupImagesResponse200 | None:
    """Cleanup unused container images

     Cleans up unused container images in the workspace registry. Only removes images that are not
    currently referenced by any active agent, function, sandbox, or job deployment.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        CleanupImagesResponse200
    """

    return sync_detailed(
        client=client,
    ).parsed


async def asyncio_detailed(
    *,
    client: Client,
) -> Response[CleanupImagesResponse200]:
    """Cleanup unused container images

     Cleans up unused container images in the workspace registry. Only removes images that are not
    currently referenced by any active agent, function, sandbox, or job deployment.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[CleanupImagesResponse200]
    """

    kwargs = _get_kwargs()

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: Client,
) -> CleanupImagesResponse200 | None:
    """Cleanup unused container images

     Cleans up unused container images in the workspace registry. Only removes images that are not
    currently referenced by any active agent, function, sandbox, or job deployment.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        CleanupImagesResponse200
    """

    return (
        await asyncio_detailed(
            client=client,
        )
    ).parsed



import atexit
import json
import logging
import sys
import threading
import traceback
import uuid
from asyncio import CancelledError
from datetime import datetime, timezone
from typing import Any
from urllib.parse import urlparse

import httpx

from .settings import settings

logger = logging.getLogger(__name__)

# Lightweight Sentry client using httpx - only captures SDK errors
_sentry_initialized = False
_captured_exceptions: set = set()  # Track already captured exceptions to avoid duplicates

# Parsed DSN components
_sentry_config: dict[str, str] | None = None

# Queue for pending events
_pending_events: list[dict[str, Any]] = []
_flush_lock = threading.Lock()
_handlers_registered = False

# Exceptions that are part of normal control flow and should not be captured
_IGNORED_EXCEPTIONS = (
    StopIteration,  # Iterator exhaustion
    StopAsyncIteration,  # Async iterator exhaustion
    GeneratorExit,  # Generator cleanup
    KeyboardInterrupt,  # User interrupt (Ctrl+C)
    SystemExit,  # Program exit
    CancelledError,  # Async task cancellation
)

# Optional dependencies that may not be installed - import errors for these are expected
_OPTIONAL_DEPENDENCIES = ("opentelemetry",)

# SDK path patterns to identify errors originating from our SDK
_SDK_PATTERNS = [
    "blaxel/",
    "blaxel\\",
    "site-packages/blaxel",
    "site-packages\\blaxel",
]


def _is_from_sdk(error: Exception) -> bool:
    """Check if an error originated from SDK code based on stack trace."""
    tb = error.__traceback__
    if not tb:
        return False

    # Walk through the traceback
    while tb:
        filename = tb.tb_frame.f_code.co_filename
        if any(pattern in filename for pattern in _SDK_PATTERNS):
            return True
        tb = tb.tb_next

    return False


def _parse_dsn(dsn: str) -> dict[str, str] | None:
    """
    Parse a Sentry DSN into its components.
    DSN format: https://{public_key}@{host}/{project_id}
    """
    try:
        parsed = urlparse(dsn)
        public_key = parsed.username
        host = parsed.hostname
        project_id = parsed.path.lstrip("/")

        if not public_key or not host or not project_id:
            return None

        return {"public_key": public_key, "host": host, "project_id": project_id}
    except Exception:
        return None


def _generate_event_id() -> str:
    """Generate a UUID v4 for event ID."""
    return uuid.uuid4().hex


def _parse_stack_trace(exc: Exception) -> list[dict[str, Any]]:
    """Parse exception traceback into Sentry-compatible frames."""
    frames: list[dict[str, Any]] = []
    tb = traceback.extract_tb(exc.__traceback__)

    for frame in tb:
        frames.append(
            {
                "filename": frame.filename,
                "function": frame.name or "<anonymous>",
                "lineno": frame.lineno,
                "colno": 0,
            }
        )

    return frames


def _error_to_sentry_event(error: Exception) -> dict[str, Any]:
    """Convert an Exception to a Sentry event payload."""
    frames = _parse_stack_trace(error)

    return {
        "event_id": _generate_event_id(),
        "timestamp": datetime.now(timezone.utc).timestamp(),
        "platform": "python",
        "level": "error",
        "environment": settings.env,
        "release": f"sdk-python@{settings.version}",
        "tags": {
            "blaxel.workspace": settings.workspace,
            "blaxel.version": settings.version,
            "blaxel.commit": settings.commit,
        },
        "exception": {
            "values": [
                {
                    "type": type(error).__name__,
                    "value": str(error),
                    "stacktrace": {"frames": frames},
                }
            ]
        },
    }


def _send_to_sentry(event: dict[str, Any]) -> None:
    """Send an event to Sentry using httpx."""
    if not _sentry_config:
        return

    public_key = _sentry_config["public_key"]
    host = _sentry_config["host"]
    project_id = _sentry_config["project_id"]
    envelope_url = f"https://{host}/api/{project_id}/envelope/"

    # Create envelope header
    envelope_header = json.dumps(
        {
            "event_id": event["event_id"],
            "sent_at": datetime.now(timezone.utc).isoformat(),
            "dsn": f"https://{public_key}@{host}/{project_id}",
        }
    )

    # Create item header
    item_header = json.dumps({"type": "event", "content_type": "application/json"})

    # Create envelope body
    envelope = f"{envelope_header}\n{item_header}\n{json.dumps(event)}"

    try:
        httpx.post(
            envelope_url,
            headers={
                "Content-Type": "application/x-sentry-envelope",
                "X-Sentry-Auth": f"Sentry sentry_version=7, sentry_client=blaxel-sdk/{settings.version}, sentry_key={public_key}",
            },
            content=envelope,
            timeout=5.0,
        )
    except Exception:
        # Silently fail - error reporting should never break the SDK
        pass


def _get_exception_key(exc_type, exc_value, frame) -> str:
    """Generate a unique key for an exception based on type, message, and origin."""
    exc_name = exc_type.__name__ if exc_type else "Unknown"
    exc_msg = str(exc_value) if exc_value else ""
    tb = getattr(exc_value, "__traceback__", None)
    if tb:
        while tb.tb_next:
            tb = tb.tb_next
        origin = f"{tb.tb_frame.f_code.co_filename}:{tb.tb_lineno}"
    else:
        origin = f"{frame.f_code.co_filename}:{frame.f_lineno}"
    return f"{exc_name}:{exc_msg}:{origin}"


def _is_optional_dependency_error(exc_type, exc_value) -> bool:
    """Check if the exception is an import error for an optional dependency."""
    if exc_type and issubclass(exc_type, ImportError):
        msg = str(exc_value).lower()
        return any(dep in msg for dep in _OPTIONAL_DEPENDENCIES)
    return False


def _trace_blaxel_exceptions(frame, event, arg):
    """Trace function that captures exceptions from blaxel SDK code."""
    if event == "exception":
        exc_type, exc_value, exc_tb = arg

        # Skip control flow exceptions (not actual errors)
        if exc_type and issubclass(exc_type, _IGNORED_EXCEPTIONS):
            return _trace_blaxel_exceptions

        # Skip import errors for optional dependencies (expected when not installed)
        if _is_optional_dependency_error(exc_type, exc_value):
            return _trace_blaxel_exceptions

        filename = frame.f_code.co_filename

        # Only capture if it's from blaxel in site-packages
        if "site-packages/blaxel" in filename:
            # Avoid capturing the same exception multiple times using a content-based key
            exc_key = _get_exception_key(exc_type, exc_value, frame)
            if exc_key not in _captured_exceptions:
                _captured_exceptions.add(exc_key)
                capture_exception(exc_value)
                # Clean up old exception keys to prevent memory leak
                if len(_captured_exceptions) > 1000:
                    _captured_exceptions.clear()

    return _trace_blaxel_exceptions


def init_sentry() -> None:
    """Initialize the lightweight Sentry client for SDK error tracking."""
    global _sentry_initialized, _sentry_config, _handlers_registered
    try:
        dsn = settings.sentry_dsn
        if not dsn:
            return

        # Parse DSN
        _sentry_config = _parse_dsn(dsn)
        if not _sentry_config:
            return

        # Only allow dev/prod environments
        if settings.env not in ("dev", "prod"):
            return

        _sentry_initialized = True

        # Register handlers only once
        if not _handlers_registered:
            _handlers_registered = True

            # Install trace function to automatically capture SDK exceptions
            sys.settrace(_trace_blaxel_exceptions)
            threading.settrace(_trace_blaxel_exceptions)

            # Register atexit handler to flush pending events
            atexit.register(flush_sentry)

    except Exception as e:
        logger.debug(f"Error initializing Sentry: {e}")


def capture_exception(exception: Exception | None = None) -> None:
    """Capture an exception to Sentry.
    Only errors originating from SDK code will be captured.
    """
    if not _sentry_initialized or not _sentry_config or exception is None:
        return

    try:
        # Generate unique key to prevent duplicate captures
        exc_key = f"{type(exception).__name__}:{str(exception)}"
        if exc_key in _captured_exceptions:
            return

        _captured_exceptions.add(exc_key)

        # Clean up old exception keys to prevent memory leak
        if len(_captured_exceptions) > 1000:
            _captured_exceptions.clear()

        # Convert error to Sentry event and queue it
        event = _error_to_sentry_event(exception)
        with _flush_lock:
            _pending_events.append(event)

        # Send immediately (fire and forget)
        _send_to_sentry(event)

    except Exception:
        # Silently fail - error capturing should never break the SDK
        pass


def flush_sentry(timeout: float = 2.0) -> None:
    """Flush pending Sentry events."""
    if not _sentry_initialized:
        return

    with _flush_lock:
        if not _pending_events:
            return

        events_to_send = _pending_events.copy()
        _pending_events.clear()

    # Send all pending events
    for event in events_to_send:
        try:
            _send_to_sentry(event)
        except Exception:
            # Silently fail
            pass


def is_sentry_initialized() -> bool:
    """Check if Sentry is initialized and available."""
    return _sentry_initialized



"""Sandbox-specific test helpers."""

# Re-export common helpers for convenience
from tests.helpers import (
    async_sleep,
    default_image,
    default_labels,
    default_region,
    sleep,
    unique_name,
    wait_for_sandbox_deletion,
    wait_for_volume_deletion,
)

__all__ = [
    "async_sleep",
    "default_image",
    "default_labels",
    "default_region",
    "sleep",
    "unique_name",
    "wait_for_sandbox_deletion",
    "wait_for_volume_deletion",
]



import pytest
import pytest_asyncio

from blaxel.core.sandbox import CodeInterpreter
from tests.helpers import default_labels, unique_name


@pytest.mark.asyncio(loop_scope="class")
class TestCodeInterpreterOperations:
    """Test CodeInterpreter operations."""

    interpreter: CodeInterpreter
    interpreter_name: str

    @pytest_asyncio.fixture(autouse=True, scope="class", loop_scope="class")
    async def setup_interpreter(self, request):
        """Set up a code interpreter for the test class."""
        request.cls.interpreter_name = unique_name("interp")
        request.cls.interpreter = await CodeInterpreter.create(
            {
                "name": request.cls.interpreter_name,
                "labels": default_labels,
            }
        )

        yield

        # Cleanup
        if request.cls.interpreter and request.cls.interpreter.metadata:
            try:
                await self.interpreter.delete()
            except Exception:
                pass


@pytest.mark.asyncio(loop_scope="class")
class TestCodeInterpreterCreate(TestCodeInterpreterOperations):
    """Test interpreter creation."""

    async def test_creates_code_interpreter(self):
        """Test that a code interpreter is created."""
        assert self.interpreter.metadata is not None
        assert self.interpreter.metadata.name is not None


@pytest.mark.asyncio(loop_scope="class")
class TestCodeInterpreterContext(TestCodeInterpreterOperations):
    """Test code context creation."""

    async def test_creates_python_code_context(self):
        """Test creating a Python code context."""
        ctx = await self.interpreter.create_code_context(language="python")

        assert ctx is not None
        assert ctx.id is not None


@pytest.mark.asyncio(loop_scope="class")
class TestRunCode(TestCodeInterpreterOperations):
    """Test runCode operations."""

    async def test_executes_simple_python_code(self):
        """Test executing simple Python code."""
        stdout_lines = []

        await self.interpreter.run_code(
            "print('Hello from interpreter')",
            language="python",
            on_stdout=lambda msg: stdout_lines.append(msg.text),
            timeout=30.0,
        )

        assert "Hello from interpreter" in "".join(stdout_lines)

    async def test_captures_stderr_output(self):
        """Test capturing stderr output."""
        stderr_lines = []

        await self.interpreter.run_code(
            "import sys; sys.stderr.write('error message')",
            language="python",
            on_stderr=lambda msg: stderr_lines.append(msg.text),
            timeout=30.0,
        )

        assert "error message" in "".join(stderr_lines)

    async def test_returns_execution_results(self):
        """Test returning execution results."""
        results = []

        await self.interpreter.run_code(
            "2 + 2",
            language="python",
            on_result=lambda res: results.append(res),
            timeout=30.0,
        )

        assert len(results) > 0

    async def test_captures_execution_errors(self):
        """Test capturing execution errors."""
        errors = []

        await self.interpreter.run_code(
            "raise ValueError('test error')",
            language="python",
            on_error=lambda err: errors.append(err),
            timeout=30.0,
        )

        assert len(errors) > 0
        assert errors[0].name == "ValueError"

    async def test_persists_state_across_runs(self):
        """Test that state persists across runs."""
        # Define a function in first run
        await self.interpreter.run_code(
            "def add(a, b):\n    return a + b",
            language="python",
            timeout=30.0,
        )

        # Call the function in second run
        stdout_lines = []
        await self.interpreter.run_code(
            "print(add(2, 3))",
            language="python",
            on_stdout=lambda msg: stdout_lines.append(msg.text),
            timeout=30.0,
        )

        assert "5" in "".join(stdout_lines)

    async def test_handles_variables_across_runs(self):
        """Test handling variables across runs."""
        # Set variable
        await self.interpreter.run_code(
            "x = 42",
            language="python",
            timeout=30.0,
        )

        # Read variable
        stdout_lines = []
        await self.interpreter.run_code(
            "print(x)",
            language="python",
            on_stdout=lambda msg: stdout_lines.append(msg.text),
            timeout=30.0,
        )

        assert "42" in "".join(stdout_lines)


@pytest.mark.asyncio(loop_scope="class")
class TestCodeInterpreterStaticMethods(TestCodeInterpreterOperations):
    """Test static methods."""

    async def test_gets_existing_interpreter(self):
        """Test getting an existing interpreter."""
        retrieved = await CodeInterpreter.get(self.interpreter.metadata.name)

        assert retrieved.metadata.name == self.interpreter.metadata.name



import pytest
import pytest_asyncio

from blaxel.core.sandbox import SandboxInstance
from tests.helpers import async_sleep, default_image, default_labels, unique_name


@pytest.mark.asyncio(loop_scope="class")
class TestProcessOperations:
    """Test sandbox process operations."""

    sandbox: SandboxInstance
    sandbox_name: str

    @pytest_asyncio.fixture(autouse=True, scope="class", loop_scope="class")
    async def setup_sandbox(self, request):
        """Set up a sandbox for the test class."""
        request.cls.sandbox_name = unique_name("process-test")
        request.cls.sandbox = await SandboxInstance.create(
            {
                "name": request.cls.sandbox_name,
                "image": default_image,
                "memory": 2048,
                "labels": default_labels,
            }
        )

        yield

        # Cleanup
        try:
            await self.sandbox.delete()
        except Exception:
            pass


@pytest.mark.asyncio(loop_scope="class")
class TestExec(TestProcessOperations):
    """Test exec operations."""

    async def test_executes_simple_command(self):
        """Test executing a simple command."""
        result = await self.sandbox.process.exec(
            {
                "command": "echo 'Hello World'",
                "wait_for_completion": True,
            }
        )

        assert result.status == "completed"
        assert "Hello World" in result.logs

    async def test_executes_command_with_custom_name(self):
        """Test executing a command with a custom name."""
        result = await self.sandbox.process.exec(
            {
                "name": "custom-named-process",
                "command": "echo 'named'",
                "wait_for_completion": True,
            }
        )

        assert result.name == "custom-named-process"

    async def test_generates_name_when_not_provided(self):
        """Test that a name is generated when not provided."""
        result = await self.sandbox.process.exec(
            {
                "command": "echo 'auto'",
                "wait_for_completion": True,
            }
        )

        assert result.name is not None
        assert result.name.startswith("proc-")

    async def test_executes_command_with_working_directory(self):
        """Test executing a command with a working directory."""
        await self.sandbox.fs.mkdir("/tmp/workdir")

        result = await self.sandbox.process.exec(
            {
                "command": "pwd",
                "working_dir": "/tmp/workdir",
                "wait_for_completion": True,
            }
        )

        assert "/tmp/workdir" in result.logs

    async def test_captures_stdout(self):
        """Test capturing stdout."""
        result = await self.sandbox.process.exec(
            {
                "command": "echo 'stdout output'",
                "wait_for_completion": True,
            }
        )

        assert "stdout output" in result.logs

    async def test_captures_stderr(self):
        """Test capturing stderr."""
        result = await self.sandbox.process.exec(
            {
                "command": "echo 'stderr output' >&2",
                "wait_for_completion": True,
            }
        )

        assert "stderr output" in result.logs

    async def test_returns_exit_code(self):
        """Test returning exit codes."""
        success_result = await self.sandbox.process.exec(
            {
                "command": "exit 0",
                "wait_for_completion": True,
            }
        )
        assert success_result.exit_code == 0

        fail_result = await self.sandbox.process.exec(
            {
                "command": "exit 42",
                "wait_for_completion": True,
            }
        )
        assert fail_result.exit_code == 42


@pytest.mark.asyncio(loop_scope="class")
class TestExecWithOnLogCallback(TestProcessOperations):
    """Test exec with onLog callback."""

    async def test_receives_logs_via_callback(self):
        """Test receiving logs via callback."""
        logs = []

        await self.sandbox.process.exec(
            {
                "command": "echo 'line1' && echo 'line2' && echo 'line3'",
                "wait_for_completion": True,
                "on_log": lambda log: logs.append(log),
            }
        )

        assert len(logs) > 0
        all_logs = " ".join(logs)
        assert "line1" in all_logs
        assert "line2" in all_logs
        assert "line3" in all_logs


@pytest.mark.asyncio(loop_scope="class")
class TestExecWithoutWaiting(TestProcessOperations):
    """Test exec without waiting for completion."""

    async def test_returns_immediately_when_wait_for_completion_is_false(self):
        """Test that exec returns immediately when waitForCompletion is false."""
        import time

        start_time = time.time()

        result = await self.sandbox.process.exec(
            {
                "name": "no-wait-test",
                "command": "sleep 5",
                "wait_for_completion": False,
            }
        )

        elapsed = time.time() - start_time
        assert elapsed < 4  # Should return well before 5 seconds
        assert result.name == "no-wait-test"

        # Cleanup
        await self.sandbox.process.kill("no-wait-test")


@pytest.mark.asyncio(loop_scope="class")
class TestProcessGet(TestProcessOperations):
    """Test process get operations."""

    async def test_retrieves_process_information(self):
        """Test retrieving process information."""
        await self.sandbox.process.exec(
            {
                "name": "get-test",
                "command": "echo 'test'",
                "wait_for_completion": True,
            }
        )

        process = await self.sandbox.process.get("get-test")

        assert process.name == "get-test"
        assert process.status == "completed"

    async def test_shows_running_status_for_long_process(self):
        """Test showing running status for long process."""
        await self.sandbox.process.exec(
            {
                "name": "long-running",
                "command": "sleep 30",
                "wait_for_completion": False,
            }
        )

        process = await self.sandbox.process.get("long-running")
        assert process.status == "running"

        # Clean up
        await self.sandbox.process.kill("long-running")


@pytest.mark.asyncio(loop_scope="class")
class TestProcessLogs(TestProcessOperations):
    """Test process logs operations."""

    async def test_retrieves_all_logs(self):
        """Test retrieving all logs."""
        await self.sandbox.process.exec(
            {
                "name": "logs-test",
                "command": "echo 'stdout' && echo 'stderr' >&2",
                "wait_for_completion": True,
            }
        )

        logs = await self.sandbox.process.logs("logs-test", "all")

        assert "stdout" in logs
        assert "stderr" in logs

    async def test_retrieves_stdout_only(self):
        """Test retrieving stdout only."""
        await self.sandbox.process.exec(
            {
                "name": "stdout-only",
                "command": "echo 'out' && echo 'err' >&2",
                "wait_for_completion": True,
            }
        )

        logs = await self.sandbox.process.logs("stdout-only", "stdout")
        assert "out" in logs

    async def test_retrieves_stderr_only(self):
        """Test retrieving stderr only."""
        await self.sandbox.process.exec(
            {
                "name": "stderr-only",
                "command": "echo 'out' && echo 'err' >&2",
                "wait_for_completion": True,
            }
        )

        logs = await self.sandbox.process.logs("stderr-only", "stderr")
        assert "err" in logs


@pytest.mark.asyncio(loop_scope="class")
class TestStreamLogs(TestProcessOperations):
    """Test streamLogs operations."""

    async def test_streams_logs_in_real_time(self):
        """Test streaming logs in real-time."""
        logs = []

        await self.sandbox.process.exec(
            {
                "name": "stream-test",
                "command": 'for i in 1 2 3; do echo "msg $i"; sleep 1; done',
                "wait_for_completion": False,
            }
        )

        stream = self.sandbox.process.stream_logs(
            "stream-test",
            {
                "on_log": lambda log: logs.append(log),
            },
        )

        try:
            await self.sandbox.process.wait("stream-test")
            await async_sleep(1)
            assert len(logs) > 0
        finally:
            stream.close()

    async def test_can_close_stream_early(self):
        """Test that stream can be closed early."""
        logs = []

        await self.sandbox.process.exec(
            {
                "name": "close-early",
                "command": "for i in $(seq 1 10); do echo $i; sleep 1; done",
                "wait_for_completion": False,
            }
        )

        stream = self.sandbox.process.stream_logs(
            "close-early",
            {
                "on_log": lambda log: logs.append(log),
            },
        )

        try:
            await async_sleep(2)
            stream.close()

            logs_at_close = len(logs)
            await async_sleep(3)

            # No new logs should arrive after close
            assert len(logs) == logs_at_close
        finally:
            # Ensure stream is closed even if test fails
            stream.close()  # Safe to call multiple times

        # Clean up
        await self.sandbox.process.kill("close-early")


@pytest.mark.asyncio(loop_scope="class")
class TestProcessWait(TestProcessOperations):
    """Test process wait operations."""

    async def test_waits_for_process_completion(self):
        """Test waiting for process completion."""
        await self.sandbox.process.exec(
            {
                "name": "wait-test",
                "command": "sleep 2 && echo 'done'",
                "wait_for_completion": False,
            }
        )

        await self.sandbox.process.wait("wait-test")

        process = await self.sandbox.process.get("wait-test")
        assert process.status == "completed"

    async def test_respects_max_wait_timeout(self):
        """Test that maxWait timeout is respected."""
        await self.sandbox.process.exec(
            {
                "name": "timeout-test",
                "command": "sleep 60",
                "wait_for_completion": False,
            }
        )

        with pytest.raises(Exception, match="Process did not finish in time"):
            await self.sandbox.process.wait("timeout-test", max_wait=2000)

        # Cleanup
        await self.sandbox.process.kill("timeout-test")


@pytest.mark.asyncio(loop_scope="class")
class TestProcessKill(TestProcessOperations):
    """Test process kill operations."""

    async def test_kills_running_process(self):
        """Test killing a running process."""
        await self.sandbox.process.exec(
            {
                "name": "kill-test",
                "command": "sleep 60",
                "wait_for_completion": False,
            }
        )

        process = await self.sandbox.process.get("kill-test")
        assert process.status == "running"

        await self.sandbox.process.kill("kill-test")
        await async_sleep(1)

        process = await self.sandbox.process.get("kill-test")
        assert process.status in ["killed", "failed", "completed"]

    async def test_handles_killing_completed_process_gracefully(self):
        """Test that killing a completed process is handled gracefully."""
        await self.sandbox.process.exec(
            {
                "name": "already-done",
                "command": "echo 'done'",
                "wait_for_completion": True,
            }
        )

        # Should not throw
        try:
            await self.sandbox.process.kill("already-done")
        except Exception:
            # Expected - some implementations throw for already completed processes
            pass


@pytest.mark.asyncio(loop_scope="class")
class TestProcessStop(TestProcessOperations):
    """Test process stop operations."""

    async def test_stops_running_process(self):
        """Test stopping a running process gracefully."""
        await self.sandbox.process.exec(
            {
                "name": "stop-test",
                "command": "sleep 60",
                "wait_for_completion": False,
            }
        )

        process = await self.sandbox.process.get("stop-test")
        assert process.status == "running"

        await self.sandbox.process.stop("stop-test")
        await async_sleep(1)

        process = await self.sandbox.process.get("stop-test")
        assert process.status in ["stopped", "killed", "failed", "completed"]

    async def test_handles_stopping_completed_process_gracefully(self):
        """Test that stopping a completed process is handled gracefully."""
        await self.sandbox.process.exec(
            {
                "name": "stop-already-done",
                "command": "echo 'done'",
                "wait_for_completion": True,
            }
        )

        # Should not throw
        try:
            await self.sandbox.process.stop("stop-already-done")
        except Exception:
            # Expected - some implementations throw for already completed processes
            pass


@pytest.mark.asyncio(loop_scope="class")
class TestProcessList(TestProcessOperations):
    """Test process list operations."""

    async def test_lists_all_processes(self):
        """Test listing all processes."""
        # Create a few processes
        await self.sandbox.process.exec(
            {
                "name": "list-test-1",
                "command": "echo 'process 1'",
                "wait_for_completion": True,
            }
        )
        await self.sandbox.process.exec(
            {
                "name": "list-test-2",
                "command": "echo 'process 2'",
                "wait_for_completion": True,
            }
        )

        processes = await self.sandbox.process.list()

        assert processes is not None
        assert isinstance(processes, list)
        assert len(processes) >= 2

        names = [p.name for p in processes]
        assert "list-test-1" in names
        assert "list-test-2" in names

    async def test_list_includes_running_processes(self):
        """Test that list includes running processes."""
        await self.sandbox.process.exec(
            {
                "name": "list-running-test",
                "command": "sleep 30",
                "wait_for_completion": False,
            }
        )

        processes = await self.sandbox.process.list()

        running_process = next((p for p in processes if p.name == "list-running-test"), None)
        assert running_process is not None
        assert running_process.status == "running"

        # Cleanup
        await self.sandbox.process.kill("list-running-test")

    async def test_list_returns_process_details(self):
        """Test that list returns process details."""
        await self.sandbox.process.exec(
            {
                "name": "list-details-test",
                "command": "echo 'details'",
                "wait_for_completion": True,
            }
        )

        processes = await self.sandbox.process.list()

        process = next((p for p in processes if p.name == "list-details-test"), None)
        assert process is not None
        assert process.name == "list-details-test"
        assert process.pid is not None
        assert process.status is not None


@pytest.mark.asyncio(loop_scope="class")
class TestRestartOnFailure(TestProcessOperations):
    """Test restartOnFailure operations."""

    async def test_restarts_process_on_failure(self):
        """Test that process restarts on failure."""
        result = await self.sandbox.process.exec(
            {
                "name": "restart-test",
                "command": "exit 1",
                "restart_on_failure": True,
                "max_restarts": 3,
                "wait_for_completion": True,
            }
        )

        assert type(result.restart_count) == int and result.restart_count > 0
        assert type(result.restart_count) == int and result.restart_count <= 3



from .autoload import autoload
from .env import env
from .internal import get_alphanumeric_limited_hash, get_global_unique_hash
from .sentry import capture_exception, flush_sentry, init_sentry, is_sentry_initialized
from .settings import Settings, settings
from .webhook import (
    AsyncSidecarCallback,
    RequestLike,
    verify_webhook_from_request,
    verify_webhook_signature,
)

__all__ = [
    "autoload",
    "capture_exception",
    "flush_sentry",
    "init_sentry",
    "is_sentry_initialized",
    "Settings",
    "settings",
    "env",
    "get_alphanumeric_limited_hash",
    "get_global_unique_hash",
    "verify_webhook_signature",
    "verify_webhook_from_request",
    "AsyncSidecarCallback",
    "RequestLike",
]



from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.status import Status
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.agent_spec import AgentSpec
    from ..models.core_event import CoreEvent
    from ..models.metadata import Metadata


T = TypeVar("T", bound="Agent")


@_attrs_define
class Agent:
    """Serverless AI agent deployment that runs your custom agent code as an auto-scaling API endpoint. Agents are deployed
    from your code repository and expose a global inference URL for querying.

        Attributes:
            metadata (Metadata): Common metadata fields shared by all Blaxel resources including name, labels, timestamps,
                and ownership information
            spec (AgentSpec): Configuration for an AI agent including runtime settings, repository source, and deployment
                triggers
            events (Union[Unset, list['CoreEvent']]): Events happening on a resource deployed on Blaxel
            status (Union[Unset, Status]): Deployment status of a resource deployed on Blaxel
    """

    metadata: "Metadata"
    spec: "AgentSpec"
    events: Union[Unset, list["CoreEvent"]] = UNSET
    status: Union[Unset, Status] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        if type(self.metadata) is dict:
            metadata = self.metadata
        else:
            metadata = self.metadata.to_dict()

        if type(self.spec) is dict:
            spec = self.spec
        else:
            spec = self.spec.to_dict()

        events: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.events, Unset):
            events = []
            for componentsschemas_core_events_item_data in self.events:
                if type(componentsschemas_core_events_item_data) is dict:
                    componentsschemas_core_events_item = componentsschemas_core_events_item_data
                else:
                    componentsschemas_core_events_item = (
                        componentsschemas_core_events_item_data.to_dict()
                    )
                events.append(componentsschemas_core_events_item)

        status: Union[Unset, str] = UNSET
        if not isinstance(self.status, Unset):
            status = self.status.value

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "metadata": metadata,
                "spec": spec,
            }
        )
        if events is not UNSET:
            field_dict["events"] = events
        if status is not UNSET:
            field_dict["status"] = status

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T | None:
        from ..models.agent_spec import AgentSpec
        from ..models.core_event import CoreEvent
        from ..models.metadata import Metadata

        if not src_dict:
            return None
        d = src_dict.copy()
        metadata = Metadata.from_dict(d.pop("metadata"))

        spec = AgentSpec.from_dict(d.pop("spec"))

        events = []
        _events = d.pop("events", UNSET)
        for componentsschemas_core_events_item_data in _events or []:
            componentsschemas_core_events_item = CoreEvent.from_dict(
                componentsschemas_core_events_item_data
            )

            events.append(componentsschemas_core_events_item)

        _status = d.pop("status", UNSET)
        status: Union[Unset, Status]
        if isinstance(_status, Unset):
            status = UNSET
        else:
            status = Status(_status)

        agent = cls(
            metadata=metadata,
            spec=spec,
            events=events,
            status=status,
        )

        agent.additional_properties = d
        return agent

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Union

import httpx

from ... import errors
from ...client import Client
from ...models.agent import Agent
from ...models.error import Error
from ...types import Response


def _get_kwargs(
    agent_name: str,
    *,
    body: Agent,
) -> dict[str, Any]:
    headers: dict[str, Any] = {}

    _kwargs: dict[str, Any] = {
        "method": "put",
        "url": f"/agents/{agent_name}",
    }

    if type(body) is dict:
        _body = body
    else:
        _body = body.to_dict()

    _kwargs["json"] = _body
    headers["Content-Type"] = "application/json"

    _kwargs["headers"] = headers
    return _kwargs


def _parse_response(*, client: Client, response: httpx.Response) -> Union[Agent, Error] | None:
    if response.status_code == 200:
        response_200 = Agent.from_dict(response.json())

        return response_200
    if response.status_code == 400:
        response_400 = Error.from_dict(response.json())

        return response_400
    if response.status_code == 401:
        response_401 = Error.from_dict(response.json())

        return response_401
    if response.status_code == 403:
        response_403 = Error.from_dict(response.json())

        return response_403
    if response.status_code == 404:
        response_404 = Error.from_dict(response.json())

        return response_404
    if response.status_code == 500:
        response_500 = Error.from_dict(response.json())

        return response_500
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(*, client: Client, response: httpx.Response) -> Response[Union[Agent, Error]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    agent_name: str,
    *,
    client: Client,
    body: Agent,
) -> Response[Union[Agent, Error]]:
    """Update agent

     Updates an agent's configuration and triggers a new deployment. Changes to runtime settings,
    environment variables, or scaling parameters will be applied on the next deployment.

    Args:
        agent_name (str):
        body (Agent): Serverless AI agent deployment that runs your custom agent code as an auto-
            scaling API endpoint. Agents are deployed from your code repository and expose a global
            inference URL for querying.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Agent, Error]]
    """

    kwargs = _get_kwargs(
        agent_name=agent_name,
        body=body,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    agent_name: str,
    *,
    client: Client,
    body: Agent,
) -> Union[Agent, Error] | None:
    """Update agent

     Updates an agent's configuration and triggers a new deployment. Changes to runtime settings,
    environment variables, or scaling parameters will be applied on the next deployment.

    Args:
        agent_name (str):
        body (Agent): Serverless AI agent deployment that runs your custom agent code as an auto-
            scaling API endpoint. Agents are deployed from your code repository and expose a global
            inference URL for querying.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Agent, Error]
    """

    return sync_detailed(
        agent_name=agent_name,
        client=client,
        body=body,
    ).parsed


async def asyncio_detailed(
    agent_name: str,
    *,
    client: Client,
    body: Agent,
) -> Response[Union[Agent, Error]]:
    """Update agent

     Updates an agent's configuration and triggers a new deployment. Changes to runtime settings,
    environment variables, or scaling parameters will be applied on the next deployment.

    Args:
        agent_name (str):
        body (Agent): Serverless AI agent deployment that runs your custom agent code as an auto-
            scaling API endpoint. Agents are deployed from your code repository and expose a global
            inference URL for querying.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Agent, Error]]
    """

    kwargs = _get_kwargs(
        agent_name=agent_name,
        body=body,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    agent_name: str,
    *,
    client: Client,
    body: Agent,
) -> Union[Agent, Error] | None:
    """Update agent

     Updates an agent's configuration and triggers a new deployment. Changes to runtime settings,
    environment variables, or scaling parameters will be applied on the next deployment.

    Args:
        agent_name (str):
        body (Agent): Serverless AI agent deployment that runs your custom agent code as an auto-
            scaling API endpoint. Agents are deployed from your code repository and expose a global
            inference URL for querying.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Agent, Error]
    """

    return (
        await asyncio_detailed(
            agent_name=agent_name,
            client=client,
            body=body,
        )
    ).parsed



"""LangGraph Model Integration Tests."""

pytest_plugins = []
import pytest  # noqa: E402

pytest.importorskip("langgraph", reason="langgraph not installed (install with: blaxel[langgraph])")

from blaxel.langgraph import bl_model  # noqa: E402

TEST_MODELS = [
    "sandbox-openai",
]


@pytest.mark.asyncio(loop_scope="class")
class TestBlModel:
    """Test bl_model functionality."""

    @pytest.mark.parametrize("model_name", TEST_MODELS)
    async def test_can_invoke_model(self, model_name: str):
        """Test invoking a model."""
        model = await bl_model(model_name)
        result = await model.ainvoke("Say hello in one word")

        assert result is not None
        assert result.content is not None
        assert isinstance(result.content, str)



from enum import Enum


class ExpirationPolicyType(str, Enum):
    DATE = "date"
    TTL_IDLE = "ttl-idle"
    TTL_MAX_AGE = "ttl-max-age"

    def __str__(self) -> str:
        return str(self.value)

    @classmethod
    def _missing_(cls, value: object) -> "ExpirationPolicyType | None":
        if isinstance(value, str):
            upper_value = value.upper()
            for member in cls:
                if member.value.upper() == upper_value:
                    return member
        return None



from .image import (
    SANDBOX_API_IMAGE,
    SANDBOX_API_PATH,
    ImageBuildContext,
    ImageInstance,
    LocalFile,
)

__all__ = [
    "ImageInstance",
    "ImageBuildContext",
    "LocalFile",
    "SANDBOX_API_IMAGE",
    "SANDBOX_API_PATH",
]



"""Blaxel - AI development platform SDK."""

from .core.common.autoload import autoload
from .core.common.env import env
from .core.common.settings import settings

__version__ = "0.2.44"
__commit__ = "68d786bff544ccf27c17ae5f45189645df7ceaec"
__sentry_dsn__ = "https://9711de13cd02b285ca4378c01de8dc30@o4508714045276160.ingest.us.sentry.io/4510461121462272"
__all__ = ["autoload", "settings", "env"]

autoload()



"""LiveKit Model Integration Tests."""

pytest_plugins = []
import pytest  # noqa: E402

pytest.importorskip(
    "livekit.agents", reason="livekit-agents not installed (install with: blaxel[livekit])"
)

from livekit.agents.llm import ChatContext  # noqa: E402

from blaxel.livekit import bl_model  # noqa: E402

TEST_MODELS = [
    "sandbox-openai",
]


@pytest.mark.asyncio(loop_scope="class")
class TestBlModel:
    """Test bl_model functionality."""

    @pytest.mark.parametrize("model_name", TEST_MODELS)
    async def test_can_create_model(self, model_name: str):
        """Test creating a model."""
        model = await bl_model(model_name)

        assert model is not None

    @pytest.mark.parametrize("model_name", TEST_MODELS)
    async def test_can_call_model(self, model_name: str):
        """Test making an actual request to the model."""
        model = await bl_model(model_name)

        chat_ctx = ChatContext.empty()
        chat_ctx.add_message(role="user", content="Say hello in one word")

        stream = model.chat(chat_ctx=chat_ctx)

        collected_text = ""
        async for chunk in stream:
            if chunk.delta and chunk.delta.content:
                collected_text += chunk.delta.content

        assert len(collected_text) > 0



from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING, Any, AsyncIterator, Iterator, List

from blaxel.core import bl_model as bl_model_core
from blaxel.core import settings

if TYPE_CHECKING:
    from langchain_core.callbacks import Callbacks  # type: ignore[import-not-found]
    from langchain_core.language_models import (  # type: ignore[import-not-found]
        LanguageModelInput,
    )
    from langchain_core.messages import BaseMessage  # type: ignore[import-not-found]
    from langchain_core.outputs import LLMResult  # type: ignore[import-not-found]
    from langchain_core.runnables import (  # type: ignore[import-not-found]
        RunnableConfig,
    )

logger = getLogger(__name__)


class TokenRefreshingWrapper:
    """Base wrapper class that refreshes token before each call."""

    def __init__(self, model_config: dict):
        self.model_config = model_config
        self.wrapped_model = self._create_model()

    def _create_model(self):
        """Create the model instance with current token."""
        config = self.model_config
        model_type = config["type"]
        model = config["model"]
        url = config["url"]
        kwargs = config.get("kwargs", {})

        if model_type == "mistral":
            from langchain_openai import ChatOpenAI  # type: ignore[import-not-found]

            return ChatOpenAI(
                api_key=settings.auth.token,
                model=model,
                base_url=f"{url}/v1",
                **kwargs,
            )
        elif model_type == "cohere":
            from langchain_cohere import ChatCohere  # type: ignore[import-not-found]

            return ChatCohere(
                cohere_api_key=settings.auth.token,
                model=model,
                base_url=url,
                **kwargs,
            )
        elif model_type == "xai":
            from langchain_xai import ChatXAI  # type: ignore[import-not-found]

            return ChatXAI(
                model=model,
                api_key=settings.auth.token,
                xai_api_base=f"{url}/v1",
                **kwargs,
            )
        elif model_type == "deepseek":
            from langchain_deepseek import (  # type: ignore[import-not-found]
                ChatDeepSeek,
            )

            return ChatDeepSeek(
                api_key=settings.auth.token,
                model=model,
                api_base=f"{url}/v1",
                **kwargs,
            )
        elif model_type == "anthropic":
            from langchain_anthropic import (  # type: ignore[import-not-found]
                ChatAnthropic,
            )

            return ChatAnthropic(
                api_key=settings.auth.token,
                anthropic_api_url=url,
                model=model,
                default_headers=settings.auth.get_headers(),
                **kwargs,
            )
        elif model_type == "gemini":
            from .custom.gemini import (
                ChatGoogleGenerativeAI,  # type: ignore[import-not-found]
            )

            return ChatGoogleGenerativeAI(
                model=model,
                client_options={"api_endpoint": url},
                additional_headers=settings.auth.get_headers(),
                transport="rest",
                **kwargs,
            )
        elif model_type == "cerebras":
            from langchain_cerebras import (  # type: ignore[import-not-found]
                ChatCerebras,
            )

            return ChatCerebras(
                api_key=settings.auth.token,
                model=model,
                base_url=f"{url}/v1",
                **kwargs,
            )
        else:
            from langchain_openai import ChatOpenAI  # type: ignore[import-not-found]

            if model_type != "openai":
                logger.warning(f"Model {model} is not supported by Langchain, defaulting to OpenAI")
            return ChatOpenAI(
                api_key=settings.auth.token,
                model=model,
                base_url=f"{url}/v1",
                **kwargs,
            )

    def _refresh_token(self):
        """Refresh the token and recreate the model if needed."""
        # Only refresh if using ClientCredentials (which has get_token method)
        current_token = settings.auth.token

        new_token = settings.auth.token

        # If token changed, recreate the model
        if current_token != new_token:
            self.wrapped_model = self._create_model()

    def __getattr__(self, name):
        """Delegate attribute access to wrapped model."""
        return getattr(self.wrapped_model, name)


class TokenRefreshingChatModel(TokenRefreshingWrapper):
    """Wrapper for chat models that refreshes token before each call."""

    async def ainvoke(
        self,
        input: LanguageModelInput,
        config: RunnableConfig | None = None,
        *,
        stop: List[str] | None = None,
        **kwargs: Any,
    ) -> BaseMessage:
        """Async invoke with token refresh."""
        self._refresh_token()
        return await self.wrapped_model.ainvoke(input, config, stop=stop, **kwargs)

    def invoke(
        self,
        input: LanguageModelInput,
        config: RunnableConfig | None = None,
        *,
        stop: List[str] | None = None,
        **kwargs: Any,
    ) -> BaseMessage:
        """Sync invoke with token refresh."""
        self._refresh_token()
        return self.wrapped_model.invoke(input, config, stop=stop, **kwargs)

    async def astream(
        self,
        input: LanguageModelInput,
        config: RunnableConfig | None = None,
        *,
        stop: List[str] | None = None,
        **kwargs: Any,
    ) -> AsyncIterator[BaseMessage]:
        """Async stream with token refresh."""
        self._refresh_token()
        async for chunk in self.wrapped_model.astream(input, config, stop=stop, **kwargs):
            yield chunk

    def stream(
        self,
        input: LanguageModelInput,
        config: RunnableConfig | None = None,
        *,
        stop: List[str] | None = None,
        **kwargs: Any,
    ) -> Iterator[BaseMessage]:
        """Sync stream with token refresh."""
        self._refresh_token()
        for chunk in self.wrapped_model.stream(input, config, stop=stop, **kwargs):
            yield chunk

    async def agenerate(
        self,
        messages: List[List[BaseMessage]],
        stop: List[str] | None = None,
        callbacks: Callbacks = None,
        *,
        tags: List[str] | None = None,
        metadata: dict[str, Any] | None = None,
        run_name: str | None = None,
        **kwargs: Any,
    ) -> LLMResult:
        """Async generate with token refresh."""
        self._refresh_token()
        return await self.wrapped_model.agenerate(
            messages,
            stop=stop,
            callbacks=callbacks,
            tags=tags,
            metadata=metadata,
            run_name=run_name,
            **kwargs,
        )

    def generate(
        self,
        messages: List[List[BaseMessage]],
        stop: List[str] | None = None,
        callbacks: Callbacks = None,
        *,
        tags: List[str] | None = None,
        metadata: dict[str, Any] | None = None,
        run_name: str | None = None,
        **kwargs: Any,
    ) -> LLMResult:
        """Sync generate with token refresh."""
        self._refresh_token()
        return self.wrapped_model.generate(
            messages,
            stop=stop,
            callbacks=callbacks,
            tags=tags,
            metadata=metadata,
            run_name=run_name,
            **kwargs,
        )

    async def astream_events(self, *args, **kwargs):
        """Async stream events with token refresh."""
        self._refresh_token()
        async for event in self.wrapped_model.astream_events(*args, **kwargs):
            yield event

    def batch(self, *args, **kwargs):
        """Batch with token refresh."""
        self._refresh_token()
        return self.wrapped_model.batch(*args, **kwargs)

    async def abatch(self, *args, **kwargs):
        """Async batch with token refresh."""
        self._refresh_token()
        return await self.wrapped_model.abatch(*args, **kwargs)


async def bl_model(name: str, **kwargs):
    url, type, model = await bl_model_core(name).get_parameters()

    # Store model configuration for recreation
    model_config = {"type": type, "model": model, "url": url, "kwargs": kwargs}

    # Create and return the wrapper
    return TokenRefreshingChatModel(model_config)



import threading
import time
from typing import Any, Callable, Dict, Literal, Union

import httpx

from ...common.settings import settings
from ..client.models import ProcessResponse, SuccessResponse
from ..client.models.process_request import ProcessRequest
from ..types import (
    ProcessRequestWithLog,
    ProcessResponseWithLog,
    SandboxConfiguration,
    StreamHandle,
)
from .action import SyncSandboxAction


class SyncSandboxProcess(SyncSandboxAction):
    def __init__(self, sandbox_config: SandboxConfiguration):
        super().__init__(sandbox_config)

    def stream_logs(
        self,
        process_name: str,
        options: Dict[str, Callable[[str], None]] | None = None,
    ) -> StreamHandle:
        """Stream logs from a process with automatic reconnection and deduplication.

        Returns a StreamHandle that can be used as a context manager:

            with sandbox.process.stream_logs(name, options) as handle:
                # do something
            # handle is automatically closed

        Or manually:

            handle = sandbox.process.stream_logs(name, options)
            try:
                # do something
            finally:
                handle.close()
        """
        if options is None:
            options = {}
        reconnect_interval = 30
        is_running = threading.Event()
        is_running.set()
        seen_logs = set()
        current_stream: StreamHandle | None = None
        timer_lock = threading.Lock()
        reconnect_timer: dict[str, threading.Timer | None] = {"t": None}

        def start_stream():
            nonlocal current_stream
            log_counter = [0]

            def make_dedup(cb_key: str):
                def inner(content: str):
                    key = f"{log_counter[0]}:{content}"
                    log_counter[0] += 1
                    if key not in seen_logs:
                        seen_logs.add(key)
                        if options.get(cb_key):
                            options[cb_key](content)

                return inner

            wrapped_options: Dict[str, Callable[[str], None]] = {}
            if "on_log" in options:
                wrapped_options["on_log"] = make_dedup("on_log")
            if "on_stdout" in options:
                wrapped_options["on_stdout"] = make_dedup("on_stdout")
            if "on_stderr" in options:
                wrapped_options["on_stderr"] = make_dedup("on_stderr")
            if current_stream:
                current_stream.close()
            current_stream = self._stream_logs(process_name, wrapped_options)

            def schedule():
                if is_running.is_set():
                    start_stream()

            with timer_lock:
                if reconnect_timer["t"]:
                    reconnect_timer["t"].cancel()
                t = threading.Timer(reconnect_interval, schedule)
                reconnect_timer["t"] = t
                t.daemon = True
                t.start()

        start_stream()

        def close():
            nonlocal current_stream
            is_running.clear()
            with timer_lock:
                if reconnect_timer["t"]:
                    reconnect_timer["t"].cancel()
                    reconnect_timer["t"] = None
            if current_stream:
                current_stream.close()
                current_stream = None
            seen_logs.clear()

        def wait_func(timeout=None):
            if current_stream:
                current_stream.wait(timeout=timeout)

        return StreamHandle(close, wait_func)

    def _stream_logs(
        self,
        identifier: str,
        options: Dict[str, Callable[[str], None]] | None = None,
    ) -> StreamHandle:
        """Private method to stream logs from a process with callbacks for different output types."""
        if options is None:
            options = {}
        closed = threading.Event()

        def run():
            url = f"{self.url}/process/{identifier}/logs/stream"
            headers = {**settings.headers, **self.sandbox_config.headers}
            try:
                with httpx.Client() as client_instance:
                    with client_instance.stream("GET", url, headers=headers) as response:
                        if response.status_code != 200:
                            raise Exception(f"Failed to stream logs: {response.text}")
                        buffer = ""
                        for chunk in response.iter_text():
                            if closed.is_set():
                                break
                            buffer += chunk
                            lines = buffer.split("\n")
                            buffer = lines.pop()
                            for line in lines:
                                if line.startswith("[keepalive]"):
                                    continue
                                if line.startswith("stdout:"):
                                    content = line[7:]
                                    if options.get("on_stdout"):
                                        options["on_stdout"](content)
                                    if options.get("on_log"):
                                        options["on_log"](content)
                                elif line.startswith("stderr:"):
                                    content = line[7:]
                                    if options.get("on_stderr"):
                                        options["on_stderr"](content)
                                    if options.get("on_log"):
                                        options["on_log"](content)
                                else:
                                    if options.get("on_log"):
                                        options["on_log"](line)
            except Exception as e:
                # Ignore on close
                if not closed.is_set():
                    raise e

        thread = threading.Thread(target=run, daemon=True)
        thread.start()

        def close():
            closed.set()

        def wait_func(timeout=None):
            thread.join(timeout=timeout)

        return StreamHandle(close, wait_func)

    def exec(
        self,
        process: Union[ProcessRequest, ProcessRequestWithLog, Dict[str, Any]],
    ) -> Union[ProcessResponse, ProcessResponseWithLog]:
        on_log = None
        on_stdout = None
        on_stderr = None

        if isinstance(process, ProcessRequestWithLog):
            on_log = process.on_log
            on_stdout = process.on_stdout
            on_stderr = process.on_stderr
            process = process.to_dict()

        if isinstance(process, dict):
            if "on_log" in process:
                on_log = process["on_log"]
                del process["on_log"]
            if "on_stdout" in process:
                on_stdout = process["on_stdout"]
                del process["on_stdout"]
            if "on_stderr" in process:
                on_stderr = process["on_stderr"]
                del process["on_stderr"]
            process = ProcessRequest.from_dict(process)

        should_wait_for_completion = process.wait_for_completion

        # When waiting for completion with streaming callbacks, use streaming endpoint
        if should_wait_for_completion and (on_log or on_stdout or on_stderr):
            return self._exec_with_streaming(
                process, on_log=on_log, on_stdout=on_stdout, on_stderr=on_stderr
            )
        else:
            with self.get_client() as client_instance:
                response = client_instance.post("/process", json=process.to_dict())
                response_data = None
                if response.content:
                    try:
                        response_data = response.json()
                    except Exception:
                        self.handle_response_error(response)
                        raise
                self.handle_response_error(response)
                result = ProcessResponse.from_dict(response_data)

                if on_log or on_stdout or on_stderr:
                    stream_control = self._stream_logs(
                        result.pid,
                        {"on_log": on_log, "on_stdout": on_stdout, "on_stderr": on_stderr},
                    )
                    return ProcessResponseWithLog(
                        result,
                        lambda: stream_control.close() if stream_control else None,
                    )

                return result

    def _exec_with_streaming(
        self,
        process_request: ProcessRequest,
        on_log: Callable[[str], None] | None = None,
        on_stdout: Callable[[str], None] | None = None,
        on_stderr: Callable[[str], None] | None = None,
    ) -> ProcessResponseWithLog:
        """Execute a process with streaming response handling for NDJSON."""
        import json

        headers = (
            self.sandbox_config.headers
            if self.sandbox_config.force_url
            else {**settings.headers, **self.sandbox_config.headers}
        )

        with httpx.Client() as client_instance:
            with client_instance.stream(
                "POST",
                f"{self.url}/process",
                headers={
                    **headers,
                    "Content-Type": "application/json",
                    "Accept": "text/event-stream",
                },
                json=process_request.to_dict(),
                timeout=None,
            ) as response:
                if response.status_code >= 400:
                    error_text = response.read()
                    raise Exception(f"Failed to execute process: {error_text}")

                content_type = response.headers.get("Content-Type", "")
                is_streaming = "application/x-ndjson" in content_type

                # Fallback: server doesn't support streaming, use legacy approach
                if not is_streaming:
                    content = response.read()
                    data = json.loads(content)
                    result = ProcessResponse.from_dict(data)

                    # If process already completed (server waited), emit logs through callbacks
                    if result.status == "completed" or result.status == "failed":
                        if result.stdout:
                            for line in result.stdout.split("\n"):
                                if line:
                                    if on_stdout:
                                        on_stdout(line)
                        if result.stderr:
                            for line in result.stderr.split("\n"):
                                if line:
                                    if on_stderr:
                                        on_stderr(line)
                        if result.logs:
                            for line in result.logs.split("\n"):
                                if line:
                                    if on_log:
                                        on_log(line)

                    return ProcessResponseWithLog(result, lambda: None)

                # Streaming response handling
                buffer = ""
                result = None

                for chunk in response.iter_text():
                    buffer += chunk
                    lines = buffer.split("\n")
                    buffer = lines.pop()

                    for line in lines:
                        if not line.strip():
                            continue
                        try:
                            parsed = json.loads(line)
                            parsed_type = parsed.get("type", "")
                            parsed_data = parsed.get("data", "")

                            if parsed_type == "stdout":
                                if parsed_data:
                                    if on_stdout:
                                        on_stdout(parsed_data)
                                    if on_log:
                                        on_log(parsed_data)
                            elif parsed_type == "stderr":
                                if parsed_data:
                                    if on_stderr:
                                        on_stderr(parsed_data)
                                    if on_log:
                                        on_log(parsed_data)
                            elif parsed_type == "result":
                                try:
                                    result = ProcessResponse.from_dict(json.loads(parsed_data))
                                except Exception:
                                    raise Exception(f"Failed to parse result JSON: {parsed_data}")
                        except json.JSONDecodeError:
                            continue

                # Process any remaining buffer
                if buffer.strip():
                    if buffer.startswith("result:"):
                        json_str = buffer[7:]
                        try:
                            result = ProcessResponse.from_dict(json.loads(json_str))
                        except Exception:
                            raise Exception(f"Failed to parse result JSON: {json_str}")

                if not result:
                    raise Exception("No result received from streaming response")

                return ProcessResponseWithLog(result, lambda: None)

    def wait(self, identifier: str, max_wait: int = 60000, interval: int = 1000) -> ProcessResponse:
        start_time = time.monotonic() * 1000
        status = "running"
        data = self.get(identifier)
        while status == "running":
            time.sleep(interval / 1000)
            try:
                data = self.get(identifier)
                status = data.status or "running"
            except Exception:
                break
            if (time.monotonic() * 1000) - start_time > max_wait:
                raise Exception("Process did not finish in time")
        return data

    def get(self, identifier: str) -> ProcessResponse:
        with self.get_client() as client_instance:
            response = client_instance.get(f"/process/{identifier}")
            self.handle_response_error(response)
            return ProcessResponse.from_dict(response.json())

    def list(self) -> list[ProcessResponse]:
        with self.get_client() as client_instance:
            response = client_instance.get("/process")
            self.handle_response_error(response)
            return [ProcessResponse.from_dict(item) for item in response.json()]

    def stop(self, identifier: str) -> SuccessResponse:
        with self.get_client() as client_instance:
            response = client_instance.delete(f"/process/{identifier}")
            self.handle_response_error(response)
            return SuccessResponse.from_dict(response.json())

    def kill(self, identifier: str) -> SuccessResponse:
        with self.get_client() as client_instance:
            response = client_instance.delete(f"/process/{identifier}/kill")
            self.handle_response_error(response)
            return SuccessResponse.from_dict(response.json())

    def logs(
        self,
        identifier: str,
        log_type: Literal["stdout", "stderr", "all"] = "all",
    ) -> str:
        with self.get_client() as client_instance:
            response = client_instance.get(f"/process/{identifier}/logs")
            self.handle_response_error(response)
            data = response.json()
            if log_type == "all":
                return data.get("logs", "")
            elif log_type == "stdout":
                return data.get("stdout", "")
            elif log_type == "stderr":
                return data.get("stderr", "")
            raise Exception("Unsupported log type")



import asyncio
import os

import httpx
import pytest
import pytest_asyncio

from blaxel.core.sandbox import SandboxInstance
from blaxel.core.sandbox.client.models import HealthResponse
from tests.helpers import default_labels, default_region, unique_name

# Environment-aware version
VERSION = "develop" if os.environ.get("BL_ENV") == "dev" else "latest"


async def wait_for_upgrade_complete(
    sandbox: SandboxInstance, max_wait_time: float = 30.0
) -> HealthResponse:
    """
    Wait for sandbox upgrade to complete by polling the health endpoint.
    Uses the SDK's health method which includes proper authentication.
    Returns the health data when upgrade count > 0, raises if upgrade failed.
    """
    print("[TEST] Waiting for health upgrade count > 0...")
    elapsed = 0.0
    poll_interval = 0.5
    health_data = None

    while elapsed < max_wait_time:
        try:
            health_data = await sandbox.system.health()
            upgrade_count = health_data.upgrade_count or 0
            print(
                f"[TEST] Health check - upgradeCount: {upgrade_count} (elapsed: {elapsed * 1000:.0f}ms)"
            )
            if upgrade_count > 0:
                print(f"[TEST] Upgrade completed (took {elapsed * 1000:.0f}ms)")
                return health_data
            if health_data.last_upgrade and health_data.last_upgrade.status == "failed":
                print(f"[TEST] Health check - last upgrade failed, health data: {health_data}")
                raise Exception(f"Upgrade failed: {health_data}")
        except Exception as e:
            # Re-throw upgrade failures
            if str(e).startswith("Upgrade failed:"):
                raise
            print(f"[TEST] Health check error: {e} (elapsed: {elapsed * 1000:.0f}ms)")

        await asyncio.sleep(poll_interval)
        elapsed += poll_interval

    raise Exception(
        f"Upgrade did not complete within {max_wait_time * 1000}ms. Last health data: {health_data}"
    )


class TestSystemOperations:
    """Base class for system tests with cleanup tracking."""

    created_sandboxes: list[str] = []

    @pytest_asyncio.fixture(autouse=True, scope="class", loop_scope="class")
    async def cleanup(self, request):
        """Clean up all resources after each test class."""
        request.cls.created_sandboxes = []

        yield

        # Clean up sandboxes
        await asyncio.gather(
            *[self._safe_delete_sandbox(name) for name in request.cls.created_sandboxes],
            return_exceptions=True,
        )

    async def _safe_delete_sandbox(self, name: str) -> None:
        """Safely delete a sandbox, ignoring errors."""
        try:
            await SandboxInstance.delete(name)
        except Exception:
            pass


@pytest.mark.asyncio(loop_scope="class")
class TestSystemUpgrade(TestSystemOperations):
    """Test system upgrade operations."""

    async def test_upgrades_sandbox_and_preview_remains_responsive(self):
        """Test that upgrade works and preview remains responsive."""
        name = unique_name("system-upgrade")
        print(f"[TEST] Starting test with sandbox name: {name}")

        # Create sandbox with Next.js image
        print("[TEST] Creating sandbox with blaxel/nextjs:latest image...")
        sandbox = await SandboxInstance.create(
            {
                "name": name,
                "image": "blaxel/nextjs:latest",
                "memory": 4096,
                "region": default_region,
                "ports": [{"target": 3000}],
                "labels": default_labels,
            }
        )
        self.created_sandboxes.append(name)
        print("[TEST] Sandbox created")

        sandbox_host = sandbox.metadata.url
        print(f"[TEST] Sandbox host: {sandbox_host}")
        assert sandbox_host is not None

        # Do initial health check using the SDK
        print("[TEST] Performing initial health check...")
        initial_health = await sandbox.system.health()
        print(f"[TEST] Initial health check - upgradeCount: {initial_health.upgrade_count}")

        # Start the Next.js dev server
        print("[TEST] Starting Next.js dev server...")
        await sandbox.process.exec(
            {
                "name": "nextjs-dev",
                "command": "npm run dev -- --port 3000",
                "working_dir": "/blaxel/app",
                "wait_for_ports": [3000],
            }
        )
        print("[TEST] Next.js dev server started")

        # Create a public preview on port 3000
        print("[TEST] Creating preview on port 3000...")
        preview = await sandbox.previews.create(
            {
                "metadata": {"name": "upgrade-test-preview"},
                "spec": {"port": 3000, "public": True},
            }
        )

        assert preview.spec.url is not None
        preview_url = preview.spec.url
        print(f"[TEST] Preview created with URL: {preview_url}")

        # Wait for preview to be ready and verify it's responsive
        print("[TEST] Waiting for preview to be ready...")
        preview_ready = False
        max_wait_time = 30.0
        elapsed = 0.0

        async with httpx.AsyncClient(timeout=10.0) as client:
            while elapsed < max_wait_time:
                try:
                    response = await client.get(preview_url)
                    print(
                        f"[TEST] Preview check - status: {response.status_code} (elapsed: {elapsed * 1000:.0f}ms)"
                    )
                    if response.status_code == 200:
                        preview_ready = True
                        break
                except Exception as e:
                    print(f"[TEST] Preview check - error: {e} (elapsed: {elapsed * 1000:.0f}ms)")
                await asyncio.sleep(1)
                elapsed += 1

        print(f"[TEST] Preview ready: {preview_ready} (took {elapsed * 1000:.0f}ms)")
        assert preview_ready is True

        # Verify preview is accessible before upgrade and capture content
        print("[TEST] Verifying preview is accessible before upgrade...")
        async with httpx.AsyncClient(timeout=10.0) as client:
            pre_upgrade_response = await client.get(preview_url)
        print(f"[TEST] Pre-upgrade preview status: {pre_upgrade_response.status_code}")
        assert pre_upgrade_response.status_code == 200
        pre_upgrade_content = pre_upgrade_response.text
        print(f"[TEST] Pre-upgrade preview content length: {len(pre_upgrade_content)} bytes")

        # Upgrade the sandbox system
        print(f"[TEST] Calling sandbox.system.upgrade(version={VERSION})...")
        upgrade_result = await sandbox.system.upgrade(version=VERSION)
        print(f"[TEST] Upgrade call completed, result: {upgrade_result}")
        assert upgrade_result is not None

        # Wait for health to show upgrade count > 0
        health_data = await wait_for_upgrade_complete(sandbox, max_wait_time)
        assert (health_data.upgrade_count or 0) > 0

        # Wait a bit for everything to stabilize after upgrade
        print("[TEST] Waiting 5s for stabilization...")
        await asyncio.sleep(5)

        # Verify preview URL is still responsive after upgrade
        print("[TEST] Verifying preview is still responsive after upgrade...")
        async with httpx.AsyncClient(timeout=10.0) as client:
            post_upgrade_response = await client.get(preview_url)
        print(f"[TEST] Post-upgrade preview status: {post_upgrade_response.status_code}")
        assert post_upgrade_response.status_code == 200

        # Verify we can still read content from the preview and compare sizes
        post_upgrade_content = post_upgrade_response.text
        print(f"[TEST] Post-upgrade preview content length: {len(post_upgrade_content)} bytes")
        assert post_upgrade_content is not None
        assert len(post_upgrade_content) > 0

        # Verify the content size is similar before and after upgrade (allow delta of 200 bytes)
        size_delta = abs(len(post_upgrade_content) - len(pre_upgrade_content))
        print(
            f"[TEST] Comparing content sizes - pre: {len(pre_upgrade_content)}, post: {len(post_upgrade_content)}, delta: {size_delta}"
        )
        assert size_delta <= 200

        print("[TEST] Test completed successfully!")

    async def test_upgrades_sandbox_and_running_process_persists(self):
        """Test that running process persists and completes after upgrade."""
        name = unique_name("system-upgrade-process")
        print(f"[TEST] Starting process persistence test with sandbox name: {name}")

        # Create sandbox
        print("[TEST] Creating sandbox...")
        sandbox = await SandboxInstance.create(
            {
                "name": name,
                "image": "blaxel/base-image:latest",
                "memory": 1024,
                "region": default_region,
                "labels": default_labels,
            }
        )
        self.created_sandboxes.append(name)
        print("[TEST] Sandbox created")

        # Start a sleep process that will run for 6 seconds
        sleep_duration = 6
        print(f"[TEST] Starting sleep process for {sleep_duration} seconds...")
        import time

        process_start = time.time()
        sleep_process = await sandbox.process.exec(
            {
                "name": "test-sleep",
                "command": f"sleep {sleep_duration}",
                "wait_for_completion": False,
            }
        )
        print(f"[TEST] Sleep process started with name: {sleep_process.name}")
        assert sleep_process.name == "test-sleep"

        # Wait a bit to ensure the process is running
        await asyncio.sleep(2)

        # Verify the process is running before upgrade
        print("[TEST] Checking process status before upgrade...")
        process_before_upgrade = await sandbox.process.get("test-sleep")
        print(f"[TEST] Process status before upgrade: {process_before_upgrade.status}")
        assert process_before_upgrade.status == "running"

        # Upgrade the sandbox system
        print(f"[TEST] Calling sandbox.system.upgrade(version={VERSION})...")
        upgrade_result = await sandbox.system.upgrade(version=VERSION)
        print(f"[TEST] Upgrade call completed, result: {upgrade_result}")
        assert upgrade_result is not None

        # Wait for the upgrade to complete (check health)
        health_data = await wait_for_upgrade_complete(sandbox, 10.0)
        assert (health_data.upgrade_count or 0) > 0

        # Check that the sleep process is still visible in the API after upgrade
        print("[TEST] Checking process status after upgrade...")
        process_after_upgrade = await sandbox.process.get("test-sleep")
        print(f"[TEST] Process status after upgrade: {process_after_upgrade.status}")
        assert process_after_upgrade is not None
        # The process should still be running (or completed if we took too long)
        assert process_after_upgrade.status in ["running", "completed"]

        # Calculate remaining time for the sleep to complete
        elapsed_since_start = time.time() - process_start
        expected_total_duration = sleep_duration
        remaining_time = expected_total_duration - elapsed_since_start
        print(
            f"[TEST] Elapsed since process start: {elapsed_since_start * 1000:.0f}ms, remaining: {remaining_time * 1000:.0f}ms"
        )

        # If the process is still running, wait for it to complete
        if process_after_upgrade.status == "running":
            # Wait for the process to complete with some buffer (5 seconds extra)
            wait_time = max(remaining_time + 5000, 5000)
            print(f"[TEST] Waiting {wait_time * 1000:.0f}ms for process to complete...")

            completed_process = await sandbox.process.wait(
                "test-sleep",
                max_wait=wait_time,
                interval=1,
            )
            print(
                f"[TEST] Process completed with status: {completed_process.status}, exitCode: {completed_process.exit_code}"
            )
            assert completed_process.status == "completed"
            assert completed_process.exit_code == 0

        # Verify the process completed in roughly the expected time (within 15 seconds tolerance)
        total_duration = time.time() - process_start
        print(
            f"[TEST] Total duration from process start to completion: {total_duration * 1000:.0f}ms"
        )
        print(f"[TEST] Expected duration: ~{expected_total_duration * 1000:.0f}ms")

        # The process should have completed close to the expected time
        # Allow 15 seconds tolerance for upgrade overhead
        tolerance = 15
        assert total_duration >= expected_total_duration - 2  # At least 4 seconds
        assert total_duration <= expected_total_duration + tolerance  # At most 21 seconds

        print("[TEST] Process persistence test completed successfully!")



from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.job_execution_task_status import JobExecutionTaskStatus
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.job_execution_task_condition import JobExecutionTaskCondition
    from ..models.job_execution_task_metadata import JobExecutionTaskMetadata
    from ..models.job_execution_task_spec import JobExecutionTaskSpec


T = TypeVar("T", bound="JobExecutionTask")


@_attrs_define
class JobExecutionTask:
    """Job execution task

    Attributes:
        conditions (Union[Unset, list['JobExecutionTaskCondition']]): Task conditions
        metadata (Union[Unset, JobExecutionTaskMetadata]): Job execution task metadata
        spec (Union[Unset, JobExecutionTaskSpec]): Job execution task specification
        status (Union[Unset, JobExecutionTaskStatus]): Job execution task status
    """

    conditions: Union[Unset, list["JobExecutionTaskCondition"]] = UNSET
    metadata: Union[Unset, "JobExecutionTaskMetadata"] = UNSET
    spec: Union[Unset, "JobExecutionTaskSpec"] = UNSET
    status: Union[Unset, JobExecutionTaskStatus] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        conditions: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.conditions, Unset):
            conditions = []
            for conditions_item_data in self.conditions:
                if type(conditions_item_data) is dict:
                    conditions_item = conditions_item_data
                else:
                    conditions_item = conditions_item_data.to_dict()
                conditions.append(conditions_item)

        metadata: Union[Unset, dict[str, Any]] = UNSET
        if (
            self.metadata
            and not isinstance(self.metadata, Unset)
            and not isinstance(self.metadata, dict)
        ):
            metadata = self.metadata.to_dict()
        elif self.metadata and isinstance(self.metadata, dict):
            metadata = self.metadata

        spec: Union[Unset, dict[str, Any]] = UNSET
        if self.spec and not isinstance(self.spec, Unset) and not isinstance(self.spec, dict):
            spec = self.spec.to_dict()
        elif self.spec and isinstance(self.spec, dict):
            spec = self.spec

        status: Union[Unset, str] = UNSET
        if not isinstance(self.status, Unset):
            status = self.status.value

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if conditions is not UNSET:
            field_dict["conditions"] = conditions
        if metadata is not UNSET:
            field_dict["metadata"] = metadata
        if spec is not UNSET:
            field_dict["spec"] = spec
        if status is not UNSET:
            field_dict["status"] = status

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T | None:
        from ..models.job_execution_task_condition import JobExecutionTaskCondition
        from ..models.job_execution_task_metadata import JobExecutionTaskMetadata
        from ..models.job_execution_task_spec import JobExecutionTaskSpec

        if not src_dict:
            return None
        d = src_dict.copy()
        conditions = []
        _conditions = d.pop("conditions", UNSET)
        for conditions_item_data in _conditions or []:
            conditions_item = JobExecutionTaskCondition.from_dict(conditions_item_data)

            conditions.append(conditions_item)

        _metadata = d.pop("metadata", UNSET)
        metadata: Union[Unset, JobExecutionTaskMetadata]
        if isinstance(_metadata, Unset):
            metadata = UNSET
        else:
            metadata = JobExecutionTaskMetadata.from_dict(_metadata)

        _spec = d.pop("spec", UNSET)
        spec: Union[Unset, JobExecutionTaskSpec]
        if isinstance(_spec, Unset):
            spec = UNSET
        else:
            spec = JobExecutionTaskSpec.from_dict(_spec)

        _status = d.pop("status", UNSET)
        status: Union[Unset, JobExecutionTaskStatus]
        if isinstance(_status, Unset):
            status = UNSET
        else:
            status = JobExecutionTaskStatus(_status)

        job_execution_task = cls(
            conditions=conditions,
            metadata=metadata,
            spec=spec,
            status=status,
        )

        job_execution_task.additional_properties = d
        return job_execution_task

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Union

import httpx

from ... import errors
from ...client import Client
from ...models.error import Error
from ...models.volume import Volume
from ...types import Response


def _get_kwargs(
    *,
    body: Volume,
) -> dict[str, Any]:
    headers: dict[str, Any] = {}

    _kwargs: dict[str, Any] = {
        "method": "post",
        "url": "/volumes",
    }

    if type(body) is dict:
        _body = body
    else:
        _body = body.to_dict()

    _kwargs["json"] = _body
    headers["Content-Type"] = "application/json"

    _kwargs["headers"] = headers
    return _kwargs


def _parse_response(*, client: Client, response: httpx.Response) -> Union[Error, Volume] | None:
    if response.status_code == 200:
        response_200 = Volume.from_dict(response.json())

        return response_200
    if response.status_code == 400:
        response_400 = Error.from_dict(response.json())

        return response_400
    if response.status_code == 401:
        response_401 = Error.from_dict(response.json())

        return response_401
    if response.status_code == 403:
        response_403 = Error.from_dict(response.json())

        return response_403
    if response.status_code == 409:
        response_409 = Error.from_dict(response.json())

        return response_409
    if response.status_code == 500:
        response_500 = Error.from_dict(response.json())

        return response_500
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(*, client: Client, response: httpx.Response) -> Response[Union[Error, Volume]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: Client,
    body: Volume,
) -> Response[Union[Error, Volume]]:
    """Create persistent volume

     Creates a new persistent storage volume that can be attached to sandboxes. Volumes must be created
    in a specific region and can only attach to sandboxes in the same region.

    Args:
        body (Volume): Persistent storage volume that can be attached to sandboxes for durable
            file storage across sessions. Volumes survive sandbox deletion and can be reattached to
            new sandboxes.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Error, Volume]]
    """

    kwargs = _get_kwargs(
        body=body,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: Client,
    body: Volume,
) -> Union[Error, Volume] | None:
    """Create persistent volume

     Creates a new persistent storage volume that can be attached to sandboxes. Volumes must be created
    in a specific region and can only attach to sandboxes in the same region.

    Args:
        body (Volume): Persistent storage volume that can be attached to sandboxes for durable
            file storage across sessions. Volumes survive sandbox deletion and can be reattached to
            new sandboxes.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Error, Volume]
    """

    return sync_detailed(
        client=client,
        body=body,
    ).parsed


async def asyncio_detailed(
    *,
    client: Client,
    body: Volume,
) -> Response[Union[Error, Volume]]:
    """Create persistent volume

     Creates a new persistent storage volume that can be attached to sandboxes. Volumes must be created
    in a specific region and can only attach to sandboxes in the same region.

    Args:
        body (Volume): Persistent storage volume that can be attached to sandboxes for durable
            file storage across sessions. Volumes survive sandbox deletion and can be reattached to
            new sandboxes.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Error, Volume]]
    """

    kwargs = _get_kwargs(
        body=body,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: Client,
    body: Volume,
) -> Union[Error, Volume] | None:
    """Create persistent volume

     Creates a new persistent storage volume that can be attached to sandboxes. Volumes must be created
    in a specific region and can only attach to sandboxes in the same region.

    Args:
        body (Volume): Persistent storage volume that can be attached to sandboxes for durable
            file storage across sessions. Volumes survive sandbox deletion and can be reattached to
            new sandboxes.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Error, Volume]
    """

    return (
        await asyncio_detailed(
            client=client,
            body=body,
        )
    ).parsed



from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.get_workspace_features_response_200_features import (
        GetWorkspaceFeaturesResponse200Features,
    )


T = TypeVar("T", bound="GetWorkspaceFeaturesResponse200")


@_attrs_define
class GetWorkspaceFeaturesResponse200:
    """
    Attributes:
        features (Union[Unset, GetWorkspaceFeaturesResponse200Features]): Map of feature keys to enabled state (always
            true). Disabled features are omitted.
    """

    features: Union[Unset, "GetWorkspaceFeaturesResponse200Features"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        features: Union[Unset, dict[str, Any]] = UNSET
        if (
            self.features
            and not isinstance(self.features, Unset)
            and not isinstance(self.features, dict)
        ):
            features = self.features.to_dict()
        elif self.features and isinstance(self.features, dict):
            features = self.features

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if features is not UNSET:
            field_dict["features"] = features

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T | None:
        from ..models.get_workspace_features_response_200_features import (
            GetWorkspaceFeaturesResponse200Features,
        )

        if not src_dict:
            return None
        d = src_dict.copy()
        _features = d.pop("features", UNSET)
        features: Union[Unset, GetWorkspaceFeaturesResponse200Features]
        if isinstance(_features, Unset):
            features = UNSET
        else:
            features = GetWorkspaceFeaturesResponse200Features.from_dict(_features)

        get_workspace_features_response_200 = cls(
            features=features,
        )

        get_workspace_features_response_200.additional_properties = d
        return get_workspace_features_response_200

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Test utility functions."""

import asyncio
import os
import time
import uuid

from blaxel.core.sandbox import SandboxInstance
from blaxel.core.volume import VolumeInstance

# Environment-aware configuration
env = os.environ.get("BL_ENV", "prod")
default_region = "eu-dub-1" if env == "dev" else "us-pdx-1"
default_image = "blaxel/base-image:latest"

# Default labels to identify test sandboxes in the UI
default_labels = {
    "env": "integration-test",
    "created-by": "pytest",
}


def unique_name(prefix: str = "test") -> str:
    """Generate a unique sandbox/volume name for testing."""
    return f"{prefix}-{uuid.uuid4().hex[:8]}"


async def wait_for_sandbox_deployed(sandbox_name: str, max_attempts: int = 30) -> bool:
    """
    Wait for a sandbox to be deployed by polling until status is DEPLOYED.

    Args:
        sandbox_name: The name of the sandbox to wait for
        max_attempts: Maximum number of attempts to wait (default: 30 seconds)

    Returns:
        True if deployed, False if timeout
    """
    attempts = 0

    while attempts < max_attempts:
        sandbox = await SandboxInstance.get(sandbox_name)
        if sandbox.status == "DEPLOYED":
            return True
        await async_sleep(1)
        attempts += 1

    print(f"Timeout waiting for {sandbox_name} to be deployed")
    return False


async def wait_for_sandbox_deletion(sandbox_name: str, max_attempts: int = 30) -> bool:
    """
    Wait for a sandbox deletion to fully complete by polling until the sandbox no longer exists.

    Args:
        sandbox_name: The name of the sandbox to wait for deletion
        max_attempts: Maximum number of attempts to wait (default: 30 seconds)

    Returns:
        True if deletion completed, False if timeout
    """
    attempts = 0

    while attempts < max_attempts:
        try:
            await SandboxInstance.get(sandbox_name)
            # If we get here, sandbox still exists, wait and try again
            await async_sleep(1)
            attempts += 1
        except Exception:
            # If get throws an error, the sandbox no longer exists
            return True

    print(f"Timeout waiting for {sandbox_name} deletion to complete")
    return False


async def wait_for_volume_deletion(volume_name: str, max_attempts: int = 30) -> bool:
    """
    Wait for a volume deletion to fully complete by polling until the volume no longer exists.

    Args:
        volume_name: The name of the volume to wait for deletion
        max_attempts: Maximum number of attempts to wait (default: 30 seconds)

    Returns:
        True if deletion completed, False if timeout
    """
    attempts = 0

    while attempts < max_attempts:
        try:
            await VolumeInstance.get(volume_name)
            # If we get here, volume still exists, wait and try again
            await async_sleep(1)
            attempts += 1
        except Exception:
            # If get throws an error, the volume no longer exists
            return True

    print(f"Timeout waiting for {volume_name} deletion to complete")
    return False


def sleep(seconds: float) -> None:
    """Synchronous sleep helper."""
    time.sleep(seconds)


async def async_sleep(seconds: float) -> None:
    """Async sleep helper."""
    await asyncio.sleep(seconds)



import asyncio
from typing import Any, Callable, Dict, Literal, Union

import httpx

from ...common.settings import settings
from ..client.models import ProcessResponse, SuccessResponse
from ..client.models.process_request import ProcessRequest
from ..types import (
    AsyncStreamHandle,
    ProcessRequestWithLog,
    ProcessResponseWithLog,
    SandboxConfiguration,
)
from .action import SandboxAction


class SandboxProcess(SandboxAction):
    def __init__(self, sandbox_config: SandboxConfiguration):
        super().__init__(sandbox_config)

    def stream_logs(
        self,
        process_name: str,
        options: Dict[str, Callable[[str], None]] | None = None,
    ) -> AsyncStreamHandle:
        """Stream logs from a process with automatic reconnection and deduplication.

        Returns an AsyncStreamHandle that can be used as a context manager:

            async with sandbox.process.stream_logs(name, options) as handle:
                # do something
            # handle is automatically closed

        Or manually:

            handle = sandbox.process.stream_logs(name, options)
            try:
                # do something
            finally:
                handle.close()
        """
        if options is None:
            options = {}

        reconnect_interval = 30  # 30 seconds in Python (TypeScript uses milliseconds)
        current_stream: AsyncStreamHandle | None = None
        is_running = True
        reconnect_timer = None

        # Track seen logs to avoid duplicates
        seen_logs = set()

        def start_stream():
            nonlocal current_stream, reconnect_timer
            log_counter = [0]  # Use list to make it mutable in nested function

            # Close existing stream if any
            if current_stream:
                current_stream.close()

            # Create wrapper options with deduplication
            wrapped_options = {}

            if "on_log" in options:

                def deduplicated_on_log(log: str):
                    log_key = f"{log_counter[0]}:{log}"
                    log_counter[0] += 1
                    if log_key not in seen_logs:
                        seen_logs.add(log_key)
                        options["on_log"](log)

                wrapped_options["on_log"] = deduplicated_on_log

            if "on_stdout" in options:

                def deduplicated_on_stdout(stdout: str):
                    log_key = f"{log_counter[0]}:{stdout}"
                    log_counter[0] += 1
                    if log_key not in seen_logs:
                        seen_logs.add(log_key)
                        options["on_stdout"](stdout)

                wrapped_options["on_stdout"] = deduplicated_on_stdout

            if "on_stderr" in options:

                def deduplicated_on_stderr(stderr: str):
                    log_key = f"{log_counter[0]}:{stderr}"
                    log_counter[0] += 1
                    if log_key not in seen_logs:
                        seen_logs.add(log_key)
                        options["on_stderr"](stderr)

                wrapped_options["on_stderr"] = deduplicated_on_stderr

            # Start new stream with deduplication
            current_stream = self._stream_logs(process_name, wrapped_options)

            # Schedule next reconnection
            if is_running:

                def schedule_reconnect():
                    if is_running:
                        start_stream()

                reconnect_timer = asyncio.get_event_loop().call_later(
                    reconnect_interval, schedule_reconnect
                )

        # Start the initial stream
        start_stream()

        # Return control functions
        def close():
            nonlocal is_running, reconnect_timer, current_stream
            is_running = False

            # Cancel reconnect timer
            if reconnect_timer:
                reconnect_timer.cancel()
                reconnect_timer = None

            # Close current stream
            if current_stream:
                current_stream.close()
                current_stream = None

            # Clear seen logs
            seen_logs.clear()

        async def wait_func():
            if current_stream:
                await current_stream.wait()

        return AsyncStreamHandle(close, wait_func)

    def _stream_logs(
        self,
        identifier: str,
        options: Dict[str, Callable[[str], None]] | None = None,
    ) -> AsyncStreamHandle:
        """Private method to stream logs from a process with callbacks for different output types."""
        if options is None:
            options = {}

        closed = False

        async def start_streaming():
            nonlocal closed

            url = f"{self.url}/process/{identifier}/logs/stream"
            headers = {**settings.headers, **self.sandbox_config.headers}

            try:
                async with httpx.AsyncClient() as client_instance:
                    async with client_instance.stream("GET", url, headers=headers) as response:
                        if response.status_code != 200:
                            raise Exception(f"Failed to stream logs: {await response.aread()}")

                        buffer = ""
                        async for chunk in response.aiter_text():
                            if closed:
                                break

                            buffer += chunk
                            lines = buffer.split("\n")
                            buffer = lines.pop()  # Keep incomplete line in buffer

                            for line in lines:
                                # Skip keepalive messages
                                if line.startswith("[keepalive]"):
                                    continue
                                if line.startswith("stdout:"):
                                    content = line[7:]  # Remove 'stdout:' prefix
                                    if options.get("on_stdout"):
                                        options["on_stdout"](content)
                                    if options.get("on_log"):
                                        options["on_log"](content)
                                elif line.startswith("stderr:"):
                                    content = line[7:]  # Remove 'stderr:' prefix
                                    if options.get("on_stderr"):
                                        options["on_stderr"](content)
                                    if options.get("on_log"):
                                        options["on_log"](content)
                                else:
                                    if options.get("on_log"):
                                        options["on_log"](line)
            except Exception as e:
                # Suppress AbortError when closing
                if hasattr(e, "name") and getattr(e, "name") == "AbortError":
                    pass
                else:
                    raise e

        # Start streaming in the background
        task = asyncio.create_task(start_streaming())

        def close():
            nonlocal closed
            closed = True
            task.cancel()

        async def wait_func():
            try:
                await task
            except asyncio.CancelledError:
                pass

        return AsyncStreamHandle(close, wait_func)

    async def exec(
        self,
        process: Union[ProcessRequest, ProcessRequestWithLog, Dict[str, Any]],
    ) -> Union[ProcessResponse, ProcessResponseWithLog]:
        """Execute a process in the sandbox."""
        on_log = None
        on_stdout = None
        on_stderr = None

        if isinstance(process, ProcessRequestWithLog):
            on_log = process.on_log
            on_stdout = process.on_stdout
            on_stderr = process.on_stderr
            process = process.to_dict()

        if isinstance(process, dict):
            if "on_log" in process:
                on_log = process["on_log"]
                del process["on_log"]
            if "on_stdout" in process:
                on_stdout = process["on_stdout"]
                del process["on_stdout"]
            if "on_stderr" in process:
                on_stderr = process["on_stderr"]
                del process["on_stderr"]
            tmp_process = ProcessRequest.from_dict(process)
            assert tmp_process is not None
            process = tmp_process

        # Store original wait_for_completion setting
        should_wait_for_completion = process.wait_for_completion

        # When waiting for completion with streaming callbacks, use streaming endpoint
        if should_wait_for_completion and (on_log or on_stdout or on_stderr):
            return await self._exec_with_streaming(
                process, on_log=on_log, on_stdout=on_stdout, on_stderr=on_stderr
            )
        else:
            client = self.get_client()
            response = await client.post("/process", json=process.to_dict())
            try:
                content_bytes = await response.aread()
                self.handle_response_error(response)
                import json

                if content_bytes:
                    response_data = json.loads(content_bytes)
                    result = ProcessResponse.from_dict(response_data)
                    assert result is not None
                else:
                    raise Exception("No content received from response")
            finally:
                await response.aclose()

            if on_log or on_stdout or on_stderr:
                stream_options: dict[str, Callable[[str], None]] = {}
                if on_log:
                    stream_options["on_log"] = on_log
                if on_stdout:
                    stream_options["on_stdout"] = on_stdout
                if on_stderr:
                    stream_options["on_stderr"] = on_stderr
                stream_control = self._stream_logs(result.pid, stream_options)
                return ProcessResponseWithLog(
                    result,
                    lambda: stream_control.close() if stream_control else None,
                )

            return result

    async def _exec_with_streaming(
        self,
        process_request: ProcessRequest,
        on_log: Callable[[str], None] | None = None,
        on_stdout: Callable[[str], None] | None = None,
        on_stderr: Callable[[str], None] | None = None,
    ) -> ProcessResponseWithLog:
        """Execute a process with streaming response handling for NDJSON."""
        import json

        headers = (
            self.sandbox_config.headers
            if self.sandbox_config.force_url
            else {**settings.headers, **self.sandbox_config.headers}
        )

        async with httpx.AsyncClient() as client_instance:
            async with client_instance.stream(
                "POST",
                f"{self.url}/process",
                headers={
                    **headers,
                    "Content-Type": "application/json",
                    "Accept": "text/event-stream",
                },
                json=process_request.to_dict(),
                timeout=None,
            ) as response:
                if response.status_code >= 400:
                    error_text = await response.aread()
                    raise Exception(f"Failed to execute process: {error_text}")

                content_type = response.headers.get("Content-Type", "")
                is_streaming = "application/x-ndjson" in content_type

                # Fallback: server doesn't support streaming, use legacy approach
                if not is_streaming:
                    content = await response.aread()
                    data = json.loads(content)
                    result = ProcessResponse.from_dict(data)
                    assert result is not None

                    # If process already completed (server waited), emit logs through callbacks
                    if result.status == "completed" or result.status == "failed":
                        if result.stdout:
                            for line in result.stdout.split("\n"):
                                if line:
                                    if on_stdout:
                                        on_stdout(line)
                        if result.stderr:
                            for line in result.stderr.split("\n"):
                                if line:
                                    if on_stderr:
                                        on_stderr(line)
                        if result.logs:
                            for line in result.logs.split("\n"):
                                if line:
                                    if on_log:
                                        on_log(line)

                    return ProcessResponseWithLog(result, lambda: None)

                # Streaming response handling
                buffer = ""
                result = None

                async for chunk in response.aiter_text():
                    buffer += chunk
                    lines = buffer.split("\n")
                    buffer = lines.pop()

                    for line in lines:
                        if not line.strip():
                            continue
                        try:
                            parsed = json.loads(line)
                            parsed_type = parsed.get("type", "")
                            parsed_data = parsed.get("data", "")

                            if parsed_type == "stdout":
                                if parsed_data:
                                    if on_stdout:
                                        on_stdout(parsed_data)
                                    if on_log:
                                        on_log(parsed_data)
                            elif parsed_type == "stderr":
                                if parsed_data:
                                    if on_stderr:
                                        on_stderr(parsed_data)
                                    if on_log:
                                        on_log(parsed_data)
                            elif parsed_type == "result":
                                try:
                                    result = ProcessResponse.from_dict(json.loads(parsed_data))
                                except Exception:
                                    raise Exception(f"Failed to parse result JSON: {parsed_data}")
                        except json.JSONDecodeError:
                            continue

                # Process any remaining buffer
                if buffer.strip():
                    if buffer.startswith("result:"):
                        json_str = buffer[7:]
                        try:
                            result = ProcessResponse.from_dict(json.loads(json_str))
                        except Exception:
                            raise Exception(f"Failed to parse result JSON: {json_str}")

                if not result:
                    raise Exception("No result received from streaming response")

                return ProcessResponseWithLog(result, lambda: None)

    async def wait(
        self, identifier: str, max_wait: int = 60000, interval: int = 1000
    ) -> ProcessResponse:
        """Wait for a process to complete."""
        start_time = asyncio.get_event_loop().time() * 1000  # Convert to milliseconds
        status = "running"
        data = await self.get(identifier)

        while status == "running":
            await asyncio.sleep(interval / 1000)  # Convert to seconds
            try:
                data = await self.get(identifier)
                status = data.status or "running"
            except:
                break

            if (asyncio.get_event_loop().time() * 1000) - start_time > max_wait:
                raise Exception("Process did not finish in time")

        return data

    async def get(self, identifier: str) -> ProcessResponse:
        import json

        client = self.get_client()
        response = await client.get(f"/process/{identifier}")
        try:
            data = json.loads(await response.aread())
            self.handle_response_error(response)
            result = ProcessResponse.from_dict(data)
            assert result is not None
            return result
        finally:
            await response.aclose()

    async def list(self) -> list[ProcessResponse]:
        import json

        client = self.get_client()
        response = await client.get("/process")
        try:
            data = json.loads(await response.aread())
            self.handle_response_error(response)
            results = []
            for item in data:
                result = ProcessResponse.from_dict(item)
                assert result is not None
                results.append(result)
            return results
        finally:
            await response.aclose()

    async def stop(self, identifier: str) -> SuccessResponse:
        import json

        client = self.get_client()
        response = await client.delete(f"/process/{identifier}")
        try:
            data = json.loads(await response.aread())
            self.handle_response_error(response)
            result = SuccessResponse.from_dict(data)
            assert result is not None
            return result
        finally:
            await response.aclose()

    async def kill(self, identifier: str) -> SuccessResponse:
        import json

        client = self.get_client()
        response = await client.delete(f"/process/{identifier}/kill")
        try:
            data = json.loads(await response.aread())
            self.handle_response_error(response)
            result = SuccessResponse.from_dict(data)
            assert result is not None
            return result
        finally:
            await response.aclose()

    async def logs(
        self,
        identifier: str,
        log_type: Literal["stdout", "stderr", "all"] = "all",
    ) -> str:
        import json

        client = self.get_client()
        response = await client.get(f"/process/{identifier}/logs")
        try:
            data = json.loads(await response.aread())
            self.handle_response_error(response)
            if log_type == "all":
                return data.get("logs", "")
            elif log_type == "stdout":
                return data.get("stdout", "")
            elif log_type == "stderr":
                return data.get("stderr", "")

            raise Exception("Unsupported log type")
        finally:
            await response.aclose()



from typing import TYPE_CHECKING, Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.upgrade_status import UpgradeStatus


T = TypeVar("T", bound="HealthResponse")


@_attrs_define
class HealthResponse:
    """
    Attributes:
        arch (str):  Example: amd64.
        build_time (str):  Example: 2026-01-29 17:36:52+00:00.
        git_commit (str):  Example: abc123.
        go_version (str):  Example: go1.25.0.
        last_upgrade (UpgradeStatus):
        os (str):  Example: linux.
        started_at (str):  Example: 2026-01-29 18:45:49+00:00.
        status (str):  Example: ok.
        upgrade_count (int):
        uptime (str):  Example: 1h30m.
        uptime_seconds (float):  Example: 5400.5.
        version (str):  Example: v0.1.0.
    """

    arch: str
    build_time: str
    git_commit: str
    go_version: str
    last_upgrade: "UpgradeStatus"
    os: str
    started_at: str
    status: str
    upgrade_count: int
    uptime: str
    uptime_seconds: float
    version: str
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        arch = self.arch

        build_time = self.build_time

        git_commit = self.git_commit

        go_version = self.go_version

        if type(self.last_upgrade) is dict:
            last_upgrade = self.last_upgrade
        else:
            last_upgrade = self.last_upgrade.to_dict()

        os = self.os

        started_at = self.started_at

        status = self.status

        upgrade_count = self.upgrade_count

        uptime = self.uptime

        uptime_seconds = self.uptime_seconds

        version = self.version

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "arch": arch,
                "buildTime": build_time,
                "gitCommit": git_commit,
                "goVersion": go_version,
                "lastUpgrade": last_upgrade,
                "os": os,
                "startedAt": started_at,
                "status": status,
                "upgradeCount": upgrade_count,
                "uptime": uptime,
                "uptimeSeconds": uptime_seconds,
                "version": version,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T | None:
        from ..models.upgrade_status import UpgradeStatus

        if not src_dict:
            return None
        d = src_dict.copy()
        arch = d.pop("arch")

        build_time = d.pop("buildTime") if "buildTime" in d else d.pop("build_time")

        git_commit = d.pop("gitCommit") if "gitCommit" in d else d.pop("git_commit")

        go_version = d.pop("goVersion") if "goVersion" in d else d.pop("go_version")

        last_upgrade = UpgradeStatus.from_dict(
            d.pop("lastUpgrade") if "lastUpgrade" in d else d.pop("last_upgrade")
        )

        os = d.pop("os")

        started_at = d.pop("startedAt") if "startedAt" in d else d.pop("started_at")

        status = d.pop("status")

        upgrade_count = d.pop("upgradeCount") if "upgradeCount" in d else d.pop("upgrade_count")

        uptime = d.pop("uptime")

        uptime_seconds = d.pop("uptimeSeconds") if "uptimeSeconds" in d else d.pop("uptime_seconds")

        version = d.pop("version")

        health_response = cls(
            arch=arch,
            build_time=build_time,
            git_commit=git_commit,
            go_version=go_version,
            last_upgrade=last_upgrade,
            os=os,
            started_at=started_at,
            status=status,
            upgrade_count=upgrade_count,
            uptime=uptime,
            uptime_seconds=uptime_seconds,
            version=version,
        )

        health_response.additional_properties = d
        return health_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import logging
from contextlib import asynccontextmanager
from typing import Any
from urllib.parse import urljoin, urlparse

import anyio
import mcp.types as types
from anyio.abc import TaskStatus
from anyio.streams.memory import (
    MemoryObjectReceiveStream,
    MemoryObjectSendStream,
)
from mcp.shared.message import SessionMessage
from websockets.asyncio.client import ClientConnection
from websockets.asyncio.client import connect as ws_connect

logger = logging.getLogger(__name__)


def remove_request_params(url: str) -> str:
    return urljoin(url, urlparse(url).path)


@asynccontextmanager
async def websocket_client(
    url: str,
    headers: dict[str, Any] | None = None,
    timeout: float = 30,
):
    """
    Client transport for WebSocket.

    The `timeout` parameter controls connection timeout.
    """
    read_stream: MemoryObjectReceiveStream[SessionMessage | Exception]
    read_stream_writer: MemoryObjectSendStream[SessionMessage | Exception]

    write_stream: MemoryObjectSendStream[SessionMessage]
    write_stream_reader: MemoryObjectReceiveStream[SessionMessage]

    read_stream_writer, read_stream = anyio.create_memory_object_stream(0)
    write_stream, write_stream_reader = anyio.create_memory_object_stream(0)

    try:
        # Wrap the task group management in a try/except GeneratorExit
        try:
            async with anyio.create_task_group() as tg:
                # Convert http(s):// to ws(s)://
                ws_url = url.replace("http://", "ws://").replace("https://", "wss://")
                logger.debug(f"Connecting to WebSocket endpoint: {remove_request_params(ws_url)}")

                # Use different parameters based on websockets version
                connection_kwargs = {
                    "additional_headers": headers,
                    "open_timeout": timeout,
                }

                async with ws_connect(ws_url, **connection_kwargs) as websocket:
                    logger.debug("WebSocket connection established")

                    async def ws_reader(
                        websocket: ClientConnection,
                        task_status: TaskStatus[None] = anyio.TASK_STATUS_IGNORED,
                    ):
                        try:
                            task_status.started()
                            async for message in websocket:
                                logger.debug(f"Received WebSocket message: {message}")
                                try:
                                    parsed_message = types.JSONRPCMessage.model_validate_json(
                                        message
                                    )
                                    logger.debug(f"Received server message: {parsed_message}")
                                    await read_stream_writer.send(
                                        SessionMessage(message=parsed_message)
                                    )
                                except Exception as exc:
                                    logger.error(f"Error parsing server message: {exc}")
                                    await read_stream_writer.send(exc)
                        except Exception as exc:
                            logger.error(f"Error in ws_reader: {exc}")
                            await read_stream_writer.send(exc)
                        finally:
                            await read_stream_writer.aclose()

                    async def ws_writer(websocket: ClientConnection):
                        try:
                            async with write_stream_reader:
                                async for session_message in write_stream_reader:
                                    logger.debug(
                                        f"Sending client message: {session_message.message}"
                                    )
                                    await websocket.send(
                                        session_message.message.model_dump_json(
                                            by_alias=True,
                                            exclude_none=True,
                                        )
                                    )
                                    logger.debug("Client message sent successfully")
                        except Exception as exc:
                            logger.error(f"Error in ws_writer: {exc}")
                        finally:
                            await write_stream.aclose()

                    await tg.start(ws_reader, websocket)
                    tg.start_soon(ws_writer, websocket)

                    try:
                        yield read_stream, write_stream
                    finally:
                        # Cancel tasks when the yield returns
                        tg.cancel_scope.cancel()

        except RuntimeError:
            # Context manager is exiting via GeneratorExit, this is normal.
            # Suppress it here so it doesn't get wrapped by anyio's BaseExceptionGroup.
            # The outer finally block below will still execute for cleanup.
            pass

        # The original code had the tg.cancel_scope.cancel() inside the 'finally'
        # associated with the 'yield'. Let's ensure cancellation happens
        # reliably. It seems correctly placed already.

    finally:
        # Ensure streams are closed even if task group setup fails or GeneratorExit is caught
        await read_stream_writer.aclose()
        await write_stream.aclose()



from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="IntegrationRepository")


@_attrs_define
class IntegrationRepository:
    """Integration repository

    Attributes:
        id (Union[Unset, str]): Repository ID
        is_bl (Union[Unset, bool]): Whether the repository has Blaxel imports
        name (Union[Unset, str]): Repository name
        organization (Union[Unset, str]): Repository owner
        url (Union[Unset, str]): Repository URL
    """

    id: Union[Unset, str] = UNSET
    is_bl: Union[Unset, bool] = UNSET
    name: Union[Unset, str] = UNSET
    organization: Union[Unset, str] = UNSET
    url: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        id = self.id

        is_bl = self.is_bl

        name = self.name

        organization = self.organization

        url = self.url

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if id is not UNSET:
            field_dict["id"] = id
        if is_bl is not UNSET:
            field_dict["isBl"] = is_bl
        if name is not UNSET:
            field_dict["name"] = name
        if organization is not UNSET:
            field_dict["organization"] = organization
        if url is not UNSET:
            field_dict["url"] = url

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T | None:
        if not src_dict:
            return None
        d = src_dict.copy()
        id = d.pop("id", UNSET)

        is_bl = d.pop("isBl", d.pop("is_bl", UNSET))

        name = d.pop("name", UNSET)

        organization = d.pop("organization", UNSET)

        url = d.pop("url", UNSET)

        integration_repository = cls(
            id=id,
            is_bl=is_bl,
            name=name,
            organization=organization,
            url=url,
        )

        integration_repository.additional_properties = d
        return integration_repository

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import json
from typing import Any

from agents import FunctionTool  # type: ignore[import-not-found]
from agents.tool_context import ToolContext  # type: ignore[import-not-found]

from blaxel.core.tools import bl_tools as bl_tools_core
from blaxel.core.tools.types import Tool


def _clean_schema_for_openai(schema: dict) -> dict:
    """Clean JSON schema to be compatible with OpenAI agents SDK.

    OpenAI agents SDK doesn't allow additionalProperties on object types.
    This function recursively removes additionalProperties and $schema fields.
    """
    if not isinstance(schema, dict):
        return schema

    cleaned_schema = schema.copy()

    # Remove $schema and additionalProperties at current level
    if "$schema" in cleaned_schema:
        del cleaned_schema["$schema"]
    if "additionalProperties" in cleaned_schema:
        del cleaned_schema["additionalProperties"]

    # Ensure object type schemas have properties
    if cleaned_schema.get("type") == "object":
        if "properties" not in cleaned_schema:
            cleaned_schema["properties"] = {}
        if "required" not in cleaned_schema:
            cleaned_schema["required"] = []

    # Recursively clean properties if they exist
    if "properties" in cleaned_schema:
        cleaned_schema["properties"] = {
            k: _clean_schema_for_openai(v) for k, v in cleaned_schema["properties"].items()
        }

    # Recursively clean items for array types
    if "items" in cleaned_schema:
        cleaned_schema["items"] = _clean_schema_for_openai(cleaned_schema["items"])

    return cleaned_schema


def get_openai_tool(tool: Tool) -> FunctionTool:
    async def openai_coroutine(
        _: ToolContext[Any],
        arguments: str,
    ) -> Any:
        if not tool.coroutine:
            raise ValueError(f"Tool {tool.name} does not have a coroutine defined")
        result = await tool.coroutine(**json.loads(arguments))
        return result

    cleaned_schema = _clean_schema_for_openai(tool.input_schema)

    return FunctionTool(
        name=tool.name,
        description=tool.description,
        params_json_schema=cleaned_schema,
        on_invoke_tool=openai_coroutine,
    )


async def bl_tools(tools_names: list[str], **kwargs) -> list[FunctionTool]:
    tools = bl_tools_core(tools_names, **kwargs)
    await tools.initialize()
    return [get_openai_tool(tool) for tool in tools.get_tools()]



"""Jobs API Integration Tests.

Note: These tests require a job named "mk3" to exist in your workspace.
The job should accept tasks with a "duration" field.
"""

import pytest

from blaxel.core.client.models.create_job_execution_request import CreateJobExecutionRequest
from blaxel.core.client.models.create_job_execution_request_env import CreateJobExecutionRequestEnv
from blaxel.core.jobs import bl_job

TEST_JOB_NAME = "mk3"


class TestBlJob:
    """Test bl_job reference creation."""

    def test_can_create_a_job_reference(self):
        """Test creating a job reference."""
        job = bl_job(TEST_JOB_NAME)

        assert job is not None
        assert hasattr(job, "create_execution")
        assert hasattr(job, "get_execution")
        assert hasattr(job, "list_executions")


@pytest.mark.asyncio(loop_scope="class")
class TestJobExecutions:
    """Test job execution operations.

    Note: These tests require the job "mk3" to exist and be properly configured.
    If the job doesn't exist, tests will be skipped.
    """

    async def test_create_get_and_list_execution(self):
        """Test creating an execution, then getting its details, status, and listing executions.

        This test combines multiple operations to reduce parallel executions.
        """
        job = bl_job(TEST_JOB_NAME)

        request = CreateJobExecutionRequest(
            tasks=[{"name": "John"}],
        )
        try:
            # Create execution
            execution_id = await job.acreate_execution(request)
            assert execution_id is not None
            assert isinstance(execution_id, str)

            # Get execution details
            execution = await job.aget_execution(execution_id)
            assert execution is not None
            assert execution.status is not None

            # Get execution status
            status = await job.aget_execution_status(execution_id)
            assert status is not None
            assert isinstance(status, str)

            # List executions (should include the one we just created)
            executions = await job.alist_executions()
            assert executions is not None
            assert isinstance(executions, list)
            assert len(executions) > 0

        except Exception as e:
            if "not found" in str(e).lower() or "404" in str(e):
                pytest.skip(f"Job '{TEST_JOB_NAME}' not found in workspace")
            raise

    async def test_wait_for_execution_to_complete(self):
        """Test waiting for execution to complete."""
        job = bl_job(TEST_JOB_NAME)

        request = CreateJobExecutionRequest(
            tasks=[{"duration": 5}],
        )
        try:
            execution_id = await job.acreate_execution(request)

            completed_execution = await job.await_for_execution(
                execution_id,
                max_wait=60,  # 1 minute
                interval=3,  # 3 seconds
            )
        except Exception as e:
            if "not found" in str(e).lower() or "404" in str(e):
                pytest.skip(f"Job '{TEST_JOB_NAME}' not found or returned unexpected data")
            raise

        assert completed_execution is not None
        assert completed_execution.status in ["succeeded", "failed", "cancelled"]

    async def test_run_job_without_overrides(self):
        """Test running a job without any overrides."""
        job = bl_job(TEST_JOB_NAME)

        try:
            execution_id = await job.arun([{"name": "Richard"}, {"name": "John"}])
        except KeyError as e:
            pytest.skip(f"Job API response missing expected field: {e}")
        except Exception as e:
            if "not found" in str(e).lower() or "404" in str(e):
                pytest.skip(f"Job '{TEST_JOB_NAME}' not found in workspace")
            raise

        assert execution_id is not None
        assert isinstance(execution_id, str)

        # Verify execution was created
        execution = await job.aget_execution(execution_id)
        assert execution is not None
        assert execution.status is not None

    async def test_run_job_with_memory_override(self):
        """Test running a job with memory override."""
        job = bl_job(TEST_JOB_NAME)

        try:
            execution_id = await job.arun(tasks=[{"name": "MemoryTest"}], memory=2048)
        except KeyError as e:
            pytest.skip(f"Job API response missing expected field: {e}")
        except Exception as e:
            if "not found" in str(e).lower() or "404" in str(e):
                pytest.skip(f"Job '{TEST_JOB_NAME}' not found in workspace")
            raise

        assert execution_id is not None
        assert isinstance(execution_id, str)

        # Verify execution was created
        execution = await job.aget_execution(execution_id)
        assert execution is not None
        assert execution.status is not None

    async def test_run_job_with_env_overrides(self):
        """Test running a job with environment variable overrides."""
        job = bl_job(TEST_JOB_NAME)

        try:
            execution_id = await job.arun(
                tasks=[{"name": "EnvTest"}], env={"CUSTOM_VAR": "test_value", "DEBUG_MODE": "true"}
            )
        except KeyError as e:
            pytest.skip(f"Job API response missing expected field: {e}")
        except Exception as e:
            if "not found" in str(e).lower() or "404" in str(e):
                pytest.skip(f"Job '{TEST_JOB_NAME}' not found in workspace")
            raise

        assert execution_id is not None
        assert isinstance(execution_id, str)

        # Verify execution was created
        execution = await job.aget_execution(execution_id)
        assert execution is not None
        assert execution.status is not None

    async def test_run_job_with_both_memory_and_env_overrides(self):
        """Test running a job with both memory and env overrides."""
        job = bl_job(TEST_JOB_NAME)

        try:
            execution_id = await job.arun(
                tasks=[{"name": "CombinedTest"}],
                memory=1024,
                env={"TEST_ENV": "production", "LOG_LEVEL": "info"},
            )
        except KeyError as e:
            pytest.skip(f"Job API response missing expected field: {e}")
        except Exception as e:
            if "not found" in str(e).lower() or "404" in str(e):
                pytest.skip(f"Job '{TEST_JOB_NAME}' not found in workspace")
            raise

        assert execution_id is not None
        assert isinstance(execution_id, str)

        # Verify execution was created
        execution = await job.aget_execution(execution_id)
        assert execution is not None
        assert execution.status is not None

    async def test_create_execution_with_memory_override(self):
        """Test creating an execution with memory override.

        Memory override must be less than or equal to the job's configured memory.
        """
        job = bl_job(TEST_JOB_NAME)

        # Override memory to 512 MB (must be <= job's configured memory)
        request = CreateJobExecutionRequest(
            tasks=[{"name": "John"}],
            memory=512,
        )

        try:
            execution_id = await job.acreate_execution(request)
            assert execution_id is not None
            assert isinstance(execution_id, str)

            # Verify the execution was created
            execution = await job.aget_execution(execution_id)
            assert execution is not None
        except Exception as e:
            if "not found" in str(e).lower() or "404" in str(e):
                pytest.skip(f"Job '{TEST_JOB_NAME}' not found in workspace")
            raise

    async def test_create_execution_with_env_override(self):
        """Test creating an execution with environment variable overrides.

        Environment variables are merged with the job's configured variables.
        """
        job = bl_job(TEST_JOB_NAME)

        # Create environment overrides
        env = CreateJobExecutionRequestEnv()
        env["CUSTOM_VAR"] = "test_value"
        env["BATCH_SIZE"] = "100"

        request = CreateJobExecutionRequest(
            tasks=[{"name": "Jane"}],
            env=env,
        )

        try:
            execution_id = await job.acreate_execution(request)
            assert execution_id is not None
            assert isinstance(execution_id, str)

            # Verify the execution was created
            execution = await job.aget_execution(execution_id)
            assert execution is not None
        except Exception as e:
            if "not found" in str(e).lower() or "404" in str(e):
                pytest.skip(f"Job '{TEST_JOB_NAME}' not found in workspace")
            raise

    async def test_create_execution_with_memory_and_env_overrides(self):
        """Test creating an execution with both memory and env overrides.

        Both overrides can be applied simultaneously to a single execution.
        """
        job = bl_job(TEST_JOB_NAME)

        # Create environment overrides
        env = CreateJobExecutionRequestEnv()
        env["LOG_LEVEL"] = "debug"
        env["TIMEOUT"] = "30"

        request = CreateJobExecutionRequest(
            tasks=[{"name": "Bob"}],
            memory=1024,  # 1 GB override
            env=env,
        )

        try:
            execution_id = await job.acreate_execution(request)
            assert execution_id is not None
            assert isinstance(execution_id, str)

            # Verify the execution was created
            execution = await job.aget_execution(execution_id)
            assert execution is not None
            assert execution.status is not None
        except Exception as e:
            if "not found" in str(e).lower() or "404" in str(e):
                pytest.skip(f"Job '{TEST_JOB_NAME}' not found in workspace")
            raise



"""Volume module for persistent storage management."""

from .volume import SyncVolumeInstance, VolumeAPIError, VolumeCreateConfiguration, VolumeInstance

__all__ = ["VolumeInstance", "SyncVolumeInstance", "VolumeCreateConfiguration", "VolumeAPIError"]



from typing import TYPE_CHECKING, Any, Dict

from blaxel.core.tools import bl_tools as bl_tools_core
from blaxel.core.tools.types import Tool, ToolException

if TYPE_CHECKING:
    from langchain_core.tools import StructuredTool  # type: ignore[import-not-found]


def _clean_schema_for_openai(schema: Dict[str, Any]) -> Dict[str, Any]:
    """Clean JSON schema to be compatible with OpenAI function calling.

    OpenAI requires object schemas to have a 'properties' field, even if empty.
    This function ensures the schema is properly formatted.
    """
    if not isinstance(schema, dict):
        return schema

    cleaned = schema.copy()

    # Ensure object type schemas have properties
    if cleaned.get("type") == "object":
        if "properties" not in cleaned:
            cleaned["properties"] = {}
        if "required" not in cleaned:
            cleaned["required"] = []

    # Recursively clean nested schemas
    if "properties" in cleaned:
        cleaned["properties"] = {
            k: _clean_schema_for_openai(v) for k, v in cleaned["properties"].items()
        }
    if "items" in cleaned and isinstance(cleaned["items"], dict):
        cleaned["items"] = _clean_schema_for_openai(cleaned["items"])

    return cleaned


def get_langchain_tool(tool: Tool) -> "StructuredTool":
    from langchain_core.tools import StructuredTool  # type: ignore[import-not-found]
    from mcp.types import CallToolResult, EmbeddedResource, ImageContent, TextContent

    async def langchain_coroutine(
        **arguments: dict[str, Any],
    ) -> tuple[str | list[str], list[ImageContent | EmbeddedResource] | None]:
        if not tool.coroutine:
            raise ValueError(f"Tool {tool.name} does not have a coroutine defined")
        result: CallToolResult = await tool.coroutine(**arguments)
        text_contents: list[TextContent] = []
        non_text_contents = []
        for content in result.content:
            if isinstance(content, TextContent):
                text_contents.append(content)
            else:
                non_text_contents.append(content)

        tool_content: str | list[str] = [content.text for content in text_contents]
        if len(text_contents) == 1:
            tool_content = tool_content[0]

        if result.isError:
            raise ToolException(tool_content)

        return tool_content, non_text_contents or None

    # Clean the schema to ensure OpenAI compatibility
    cleaned_schema = _clean_schema_for_openai(tool.input_schema)

    return StructuredTool(
        name=tool.name,
        description=tool.description,
        args_schema=cleaned_schema,
        coroutine=langchain_coroutine,
        sync_coroutine=tool.sync_coroutine,
    )


async def bl_tools(tools_names: list[str], **kwargs) -> list["StructuredTool"]:
    tools = bl_tools_core(tools_names, **kwargs)
    await tools.initialize()
    return [get_langchain_tool(tool) for tool in tools.get_tools()]



"""Core blTools Integration Tests."""

import asyncio

import pytest
import pytest_asyncio

from blaxel.core.sandbox import SandboxInstance
from blaxel.core.tools import bl_tools as bl_tools_core
from tests.helpers import default_image, default_labels, unique_name


class TestCoreBlToolsSync:
    """Test core blTools sync functionality."""

    sandbox: SandboxInstance | None = None
    sandbox_name: str | None = None

    @pytest.fixture(autouse=True, scope="class")
    def setup_sandbox(self, request):
        """Set up a sandbox for the test class."""
        request.cls.sandbox_name = unique_name("core-sync-tools-test")

        async def create_sandbox():
            return await SandboxInstance.create(
                {
                    "name": request.cls.sandbox_name,
                    "image": default_image,
                    "memory": 2048,
                    "labels": default_labels,
                }
            )

        request.cls.sandbox = asyncio.get_event_loop().run_until_complete(create_sandbox())

        yield

        # Cleanup
        async def cleanup():
            try:
                await request.cls.sandbox.delete()
            except Exception:
                pass

        try:
            asyncio.get_event_loop().run_until_complete(cleanup())
        except Exception:
            pass

    def test_can_get_tool_names(self):
        """Test getting tool names."""
        tools = bl_tools_core([f"sandbox/{self.sandbox_name}"])

        assert tools.functions is not None
        assert len(tools.functions) > 0


@pytest.mark.asyncio(loop_scope="class")
class TestCoreBlTools:
    """Test core blTools functionality."""

    sandbox: SandboxInstance | None = None
    sandbox_name: str | None = None

    @pytest_asyncio.fixture(autouse=True, scope="class", loop_scope="class")
    async def setup_sandbox(self, request):
        """Set up a sandbox for the test class."""
        request.cls.sandbox_name = unique_name("core-tools-test")
        request.cls.sandbox = await SandboxInstance.create(
            {
                "name": request.cls.sandbox_name,
                "image": default_image,
                "memory": 2048,
                "labels": default_labels,
            }
        )

        yield

        # Cleanup
        try:
            await request.cls.sandbox.delete()
        except Exception:
            pass

    async def test_can_get_and_invoke_tools(self):
        """Test getting and invoking tools."""
        tools_wrapper = bl_tools_core([f"sandbox/{self.sandbox_name}"])
        await tools_wrapper.initialize()
        tools = tools_wrapper.get_tools()

        assert len(tools) > 0

        # Find the exec tool to test
        exec_tool = next((t for t in tools if "exec" in t.name.lower()), None)
        assert exec_tool is not None
        assert exec_tool.coroutine is not None
        result = await exec_tool.coroutine(
            command="echo 'hello'",
        )
        assert result is not None
